import copy

import torch
import torch.fx.experimental.optimization as optimization
import warnings

from .nn import utils
from .optim._optimizer_utils import optimizer_fusion, IPEX_FUSED_OPTIMIZER_LIST
import intel_extension_for_pytorch._C as core
from enum import IntEnum

def _copy_model_and_optimizer(model, optimizer):
    new_model = copy.deepcopy(model)
    if optimizer is None:
        return new_model, optimizer
    else:
        new_optimizer = copy.deepcopy(optimizer)
        new_optimizer.state.clear()
        dic_param = {}
        for k, value in zip(model.parameters(), new_model.parameters()):
            dic_param[k] = value
        for group1, group2 in zip(optimizer.param_groups, new_optimizer.param_groups):
            for i, p in enumerate(group1['params']):
                new_model_param = dic_param[p]
                group2['params'][i] = new_model_param
                new_optimizer.state[new_model_param] = copy.deepcopy(optimizer.state[p])
        return new_model, new_optimizer

class _Properties(object):
    r"""
    This class is to establish a set of default properties.

    """
    def __init__(self):
        self.opt_level = None
        self.conv_bn_folding = None
        self.weights_prepack = None
        self.remove_dropout = None
        # optimizer opt conig
        self.split_master_weight_for_bf16 = None
        self.fuse_update_step = None
        self.auto_kernel_selection = None

# O0 properties
class _O0:
    def __call__(self, properties):
        properties.opt_level = "O0"
        properties.conv_bn_folding = False
        properties.weights_prepack = False
        properties.replace_dropout_with_identity = False
        properties.optimize_lstm = False
        properties.split_master_weight_for_bf16 = False
        properties.fuse_update_step = False
        properties.auto_kernel_selection = False
        return properties


# O1 properties
class _O1:
    def __call__(self, properties):
        properties.opt_level = "O1"
        properties.conv_bn_folding = True
        properties.weights_prepack = True
        properties.replace_dropout_with_identity = True
        properties.optimize_lstm = True
        properties.split_master_weight_for_bf16 = True
        properties.fuse_update_step = True
        properties.auto_kernel_selection = False
        return properties

opt_levels = {"O0": _O0(),
              "O1": _O1()}

def optimize(
    model,
    dtype=None,
    optimizer=None,
    level="O1",
    inplace=False,
    conv_bn_folding=None,
    weights_prepack=None,
    replace_dropout_with_identity=None,
    optimize_lstm=None,
    split_master_weight_for_bf16=None,
    fuse_update_step=None,
    auto_kernel_selection=None,
    sample_input=None
):
    r"""
    Apply optimizations at Python frontend to the given model (nn.Module), as
    well as the given optimizer (optional). If the optimizer is given,
    optimizations will be applied for training. Otherwise, optimization will be
    applied for inference. Optimizations include ``conv+bn`` folding (for
    inference only), weight prepacking and so on.

    Weight prepacking is a technique to accelerate performance of oneDNN
    operators. In order to achieve better vectorization and cache reuse, onednn
    uses a specific memory layout called ``blocked layout``. Although the
    calculation itself with ``blocked layout`` is fast enough, from memory usage
    perspective it has drawbacks. Running with the ``blocked layout``, oneDNN
    splits one or several dimensions of data into blocks with fixed size each
    time the operator is executed. More details information about oneDNN data
    mermory format is available at `oneDNN manual <https://oneapi-src.github.io/oneDNN/dev_guide_understanding_memory_formats.html>`_.
    To reduce this overhead, data will be converted to predefined block shapes
    prior to the execution of oneDNN operator execution. In runtime, if the data
    shape matches oneDNN operator execution requirements, oneDNN won't perform
    memory layout conversion but directly go to calculation. Through this
    methodology, called ``weight prepacking``, it is possible to avoid runtime
    weight data format convertion and thus increase performance.

    Args:
        model (torch.nn.Module): User model to apply optimizations on.
        dtype (torch.dtype): Only works for ``torch.bfloat16``.
            Model parameters will be casted to ``torch.bfloat16`` if dtype is set to
            ``torch.bfloat16``. The default value is None, meaning do nothing.
            Note: Data type conversion is only applied to ``nn.Conv2d``, ``nn.Linear``
            and ``nn.ConvTranspose2d`` for both training and inference cases. For
            inference mode, additional data type conversion is applied to the weights
            of ``nn.Embedding`` and ``nn.LSTM``.
        optimizer (torch.optim.Optimizer): User optimizer to apply optimizations
            on, such as SGD. The default value is ``None``, meaning inference case.
        level (string): ``"O0"`` or ``"O1"``. No optimizations are applied with
            ``"O0"``. The optimizer function just returns the original model and
            optimizer. With ``"O1"``, the following optimizations are applied:
            conv+bn folding, weights prepack, dropout removal (inferenc model),
            master weight split and fused optimizer update step (training model).
            The optimization options can be further overridden by setting the
            following options explicitly. The default value is ``"O1"``.
        inplace (bool): Whether to perform inplace optimization. Default value is
            ``False``.
        conv_bn_folding (bool): Whether to perform ``conv_bn`` folding. It only
            works for inference model. The default value is ``None``. Explicitly
            setting this knob overwrites the configuration set by ``level`` knob.
        weights_prepack (bool): Whether to perform weight prepack for convolution
            and linear to avoid oneDNN weights reorder. The default value is
            ``None``. Explicitly setting this knob overwrites the configuration
            set by ``level`` knob.
        replace_dropout_with_identity (bool): Whether to replace ``nn.Dropout``
            with ``nn.Identity``. If replaced, the ``aten::dropout`` won't be
            included in the JIT graph. This may provide more fusion opportunites
            on the graph. This only works for inference model. The default value
            is ``None``. Explicitly setting this knob overwrites the configuration
            set by ``level`` knob.
        optimize_lstm (bool): Whether to replace ``nn.LSTM`` with ``IPEX LSTM``
            which takes advantage of oneDNN kernels to get better performance.
            The default value is ``None``. Explicitly setting this knob
            overwrites the configuration set by ``level`` knob.
        split_master_weight_for_bf16 (bool): Whether to split master weights
            update for BF16 training. This saves memory comparing to master
            weight update solution. Split master weights update methodology
            doesn't support all optimizers. The default value is None. The
            default value is ``None``. Explicitly setting this knob overwrites
            the configuration set by ``level`` knob.
        fuse_update_step (bool): Whether to use fused params update for training
            which have better performance. It doesn't support all optimizers.
            The default value is ``None``. Explicitly setting this knob
            overwrites the configuration set by ``level`` knob.
        sample_input (tuple or torch.Tensor): Whether to feed sample input data to ipex.optimize. The shape of
            input data will impact the block format of packed weight. If not feed a sample
            input, Intel® Extension for PyTorch* will pack the weight per some predefined heuristics.
            If feed a sample input with real input shape, Intel® Extension for PyTorch* can get
            best block format.            
        auto_kernel_selection (bool) [experimental]: Different backends may have
            different performances with different dtypes/shapes. Default value
            is False. Intel® Extension for PyTorch* will try to optimize the
            kernel selection for better performance if this knob is set to
            ``True``. There might be regressions at current stage. The default
            value is ``None``. Explicitly setting this knob overwrites the
            configuration set by ``level`` knob.

    Returns:
        Model and optimizer (if given) modified according to the ``level`` knob
        or other user settings. ``conv+bn`` folding may take place and
        ``dropout`` may be replaced by ``identity``. In inference scenarios,
        convolutuon, linear and lstm will be replaced with the optimized
        counterparts in Intel® Extension for PyTorch* (weight prepack for
        convolution and linear) for good performance. In bfloat16 scenarios,
        parameters of convolution and linear will be casted to bfloat16 dtype.

    .. warning::

        Please invoke ``optimize`` function AFTER loading weights to model via
        ``model.load_state_dict(torch.load(PATH))``.

    .. warning::

        Please invoke ``optimize`` function BEFORE invoking DDP in distributed
        training scenario.

        The ``optimize`` function deepcopys the original model. If DDP is invoked
        before ``optimize`` function, DDP is applied on the origin model, rather
        than the one returned from ``optimize`` function. In this case, some
        operators in DDP, like allreduce, will not be invoked and thus may cause
        unpredictable accuracy loss.

    Examples:

        >>> # bfloat16 inference case.
        >>> model = ...
        >>> model.load_state_dict(torch.load(PATH))
        >>> model.eval()
        >>> optimized_model = ipex.optimize(model, dtype=torch.bfloat16)
        >>> # running evaluation step.
        >>> # bfloat16 training case.
        >>> optimizer = ...
        >>> model.train()
        >>> optimized_model, optimized_optimizer = ipex.optimize(model, dtype=torch.bfloat16, optimizer=optimizer)
        >>> # running training step.

    """
    if isinstance(model, torch.jit.ScriptModule):
        if optimizer is None:
            return model
        return model, optimizer

    if model.training:
        assert optimizer is not None, "The optimizer should be given for training mode"
    else:
        assert optimizer is None, "The optimizer should not be given for inference mode"

    opt_properties = _Properties()
    if level not in opt_levels:
        raise RuntimeError(
            "Unexpected optimization level {}. ".format(level) +
            "Options are 'O0', 'O1'.")
    else:
        opt_properties = opt_levels[level](opt_properties)

    if level is not None:
        opt_properties.opt_level = level
    if conv_bn_folding is not None:
        opt_properties.conv_bn_folding = conv_bn_folding
    if weights_prepack is not None:
        opt_properties.weights_prepack = weights_prepack
    if replace_dropout_with_identity is not None:
        opt_properties.replace_dropout_with_identity = replace_dropout_with_identity
    if optimize_lstm is not None:
        opt_properties.optimize_lstm = optimize_lstm
    if split_master_weight_for_bf16 is not None:
        opt_properties.split_master_weight_for_bf16 = split_master_weight_for_bf16
    if fuse_update_step is not None:
        opt_properties.fuse_update_step = fuse_update_step
    if auto_kernel_selection is not None:
        opt_properties.auto_kernel_selection = auto_kernel_selection        
    if sample_input is not None:
        if isinstance(sample_input, torch.Tensor):
            sample_input = (sample_input,)
        utils._weight_prepack.record_input_shape_for_prepack(model, sample_input)

    if inplace:
        optimized_model = model
        optimized_optimizer = optimizer
    else:
        optimized_model, optimized_optimizer = _copy_model_and_optimizer(model, optimizer)

    if not model.training:
        if opt_properties.conv_bn_folding:
            try:
                optimized_model = optimization.fuse(optimized_model, inplace=inplace)
            except:  # noqa E722
                warnings.warn("Conv BatchNorm folding failed during the optimize process.")
        if opt_properties.replace_dropout_with_identity:
            utils._model_convert.replace_dropout_with_identity(optimized_model)
        if dtype == torch.bfloat16:
            optimized_model = utils._model_convert.convert_module_data_type(optimized_model, torch.bfloat16)

    if opt_properties.optimize_lstm:
        utils._model_convert.replace_lstm_with_ipex_lstm(optimized_model)
    if model.training and opt_properties.split_master_weight_for_bf16 and dtype is torch.bfloat16:
        if not opt_properties.fuse_update_step:
            opt_properties.split_master_weight_for_bf16 = False
            warnings.warn(
                "IPEX does not non-fused split master weight for bf16 training," +
                "have reset split_master_weight_for_bf16 flag to False." +
                "If you want to use split_master_weight_for_bf16." +
                "Please set both split_master_weight_for_bf16 and fuse_update_step to True")
        elif type(optimizer) not in IPEX_FUSED_OPTIMIZER_LIST:
            opt_properties.split_master_weight_for_bf16 = False
            opt_properties.fuse_update_step = False
            warnings.warn(
                "IPEX does not support fused/fused split update for" + str(type(optimizer)) +
                "will use non-fused master weight update for bf16 training")

    # convert optimizer for training case.
    params_attr = {}
    if dtype == torch.bfloat16 and model.training:
        optimized_model, optimized_optimizer, params_attr = utils._weight_cast.weight_dtype_convert_with_ipex(
            optimized_model, optimized_optimizer, params_attr, opt_properties.split_master_weight_for_bf16)
    if opt_properties.weights_prepack:
        if dtype == torch.bfloat16:
            assert core.onednn_has_bf16_support(), \
                    "BF16 weight prepack needs the cpu support avx512bw, avx512vl and avx512dq, " + \
                    "please set dtype to torch.float or set weights_prepack to False."
        optimized_model, optimized_optimizer, params_attr = utils._weight_prepack.weight_prepack_with_ipex(
            optimized_model, optimized_optimizer, params_attr, opt_properties.auto_kernel_selection)
    # TODO: model list, optimizer list.
    if optimizer is None:
        return optimized_model

    # with an optimizer
    if opt_properties.fuse_update_step:
        optimized_optimizer = optimizer_fusion(
            optimized_optimizer, opt_properties.split_master_weight_for_bf16)
    return optimized_model, optimized_optimizer


def enable_onednn_fusion(enabled):
    r"""
    Enables or disables oneDNN fusion functionality. If enabled, oneDNN
    operators will be fused in runtime, when intel_extension_for_pytorch
    is imported.

    Args:
        enabled (bool): Whether to enable oneDNN fusion functionality or not.
            Default value is ``True``.

    Examples:

        >>> import intel_extension_for_pytorch as ipex
        >>> # to enable the oneDNN fusion
        >>> ipex.enable_onednn_fusion(True)
        >>> # to disable the oneDNN fusion
        >>> ipex.enable_onednn_fusion(False)
    """

    if enabled:
        core.enable_jit_opt()
    else:
        core.disable_jit_opt()

class FP32MathMode(IntEnum):
    FP32 = int(core.FP32MathMode.FP32)
    TF32 = int(core.FP32MathMode.TF32)
    BF32 = int(core.FP32MathMode.BF32)

def set_fp32_math_mode(mode=FP32MathMode.FP32, device="cpu"):
    r"""
    Enable or disable implicit data type conversion.
    If mode is FP32MathMode.FP32 which means to disable the oneDNN fpmath mode.
    If mode is FP32MathMode.BF32 which means to enable the oneDNN fpmath mode by down convert to bfloat16 implicitly.

    Args:
        mode (FP32MathMode): Only works for ``FP32MathMode.FP32`` and ``FP32MathMode.BF32``.
            oneDNN fpmath mode will be disabled by default if dtype is set to ``FP32MathMode.FP32``.
            The implicit FP32 to BF16 data type conversion will be enabled if dtype is set to ``FP32MathMode.BF32`.
        device (string): Only "cpu" is supported right now.

    Examples:

        >>> import intel_extension_for_pytorch as ipex
        >>> # to enable the implicit data type conversion
        >>> ipex.set_fp32_math_mode(mode=ipex.FP32MathMode.BF32)
        >>> # to disable the implicit data type conversion
        >>> ipex.set_fp32_math_mode(mode=ipex.FP32MathMode.FP32)
    """

    if mode == FP32MathMode.BF32:
        core.set_fp32_math_mode(core.FP32MathMode.BF32)
    elif mode == FP32MathMode.FP32:
        core.set_fp32_math_mode(core.FP32MathMode.FP32)
    else:
        warnings.warn("IPEX does not support mode except FP32MathMode.FP32 and FP32MathMode.BF32 for fpmath_mode right now.")


def get_fp32_math_mode(device="cpu"):
    r"""
    Get the current fpmath_mode setting.

    Args:
        device (string): Only "cpu" is supported right now

    Returns:
        Fpmath mode
        The value will be ``FP32MathMode.FP32`` or ``FP32MathMode.BF32``.
        ``FP32MathMode.FP32`` means implicit down-conversion is disabled,
        while ``FP32MathMode.BF32`` means implicit down-conversions from f32 to bf16/f16 or compatible FP type is allowed.

    Examples:

        >>> import intel_extension_for_pytorch as ipex
        >>> # to get the current fpmath mode
        >>> ipex.get_fp32_math_mode(device="cpu")
    """

    return core.get_fp32_math_mode()



import copy
from typing import Tuple, Any
import warnings

import torch
from torch.ao.quantization import PlaceholderObserver
import torch.fx.experimental.optimization as optimization

import intel_extension_for_pytorch._C as core
from ._quantize_utils import auto_prepare, auto_convert, copy_prepared_model
from .. import nn

def prepare(
    model,
    configure,
    example_inputs,
    inplace=False):
    r"""
    Prepare an FP32 torch.nn.Module model to do calibration or to convert to quantized model.

    Args:
        model (torch.nn.Module): The FP32 model to be prepared.
        configure (torch.quantization.qconfig.QConfig): The observer settings about activation and weight.
        example_inputs (tuple or torch.Tensor): A tuple of example inputs that
            will be passed to the function while running to init quantization state.
        inplace: (bool): It will change the given model in-place if True. The default value is ``False``.

    Returns:
        torch.nn.Module
    """
    assert isinstance(model, torch.nn.Module), "Only support nn.Module prepare for quantization path"
    try:
        prepare_model = optimization.fuse(model, inplace=inplace)
    except:  # noqa E722
        if inplace:
            prepare_model = model
        else:
            try:
                prepare_model = copy.deepcopy(model)
            except:
                assert False, "The model's copy is failed, please try set inplace to True to do the prepare"
        warnings.warn("Conv BatchNorm folding failed during the prepare process.")
    # replace dropout with identity to enable more fusion pattern.
    nn.utils._model_convert.replace_dropout_with_identity(prepare_model)
    # Special case for common case of passing a single Tensor
    if isinstance(example_inputs, (torch.Tensor, dict)):
        example_inputs = (example_inputs,)
    elif not isinstance(example_inputs, tuple):
        example_inputs = tuple(example_inputs)
    return auto_prepare(prepare_model, configure, example_inputs)

def convert(
    model,
    inplace=False):
    r"""
    Convert an FP32 prepared model to a model which will automatically insert fake quant
    before a quantizable module or operator.

    Args:
        model (torch.nn.Module): The FP32 model to be convert.
        inplace: (bool): It will change the given model in-place if True. The default value is ``False``.

    Returns:
        torch.torch.nn.Module
    """
    assert isinstance(model, torch.nn.Module), "Only support nn.Module convert for quantization path"
    assert hasattr(model, 'q_config'), "Please do prepare the model before doing convert"

    if inplace:
        convert_model = model
    else:
        try:
            convert_model = copy_prepared_model(model)
        except:
            assert False, "The model's copy is failed, please try set inplace to True to do the convert"

    # If the module's activation's qconfig is PlaceholderObserver,
    # we can say that the module want to run dynamic quantization path.
    if isinstance(convert_model.q_config.activation(), PlaceholderObserver):
        qconfig_spec = {
            torch.nn.Linear : convert_model.q_config,
            torch.nn.LSTM : convert_model.q_config,
            torch.nn.GRU : convert_model.q_config,
            torch.nn.LSTMCell : convert_model.q_config,
            torch.nn.RNNCell : convert_model.q_config,
            torch.nn.GRUCell : convert_model.q_config,
        }
        return torch.quantization.quantize_dynamic(convert_model, qconfig_spec=qconfig_spec, inplace=True)

    # Convert linear, conv, and Embedding's weight dtype when use autocast,
    # which will reduce the dtype conversion.
    # TODO: check whether can be removed or not?
    if torch.is_autocast_cpu_enabled() and core.get_autocast_dtype() == torch.bfloat16:
        convert_model = nn.utils._model_convert.convert_module_data_type(convert_model, torch.bfloat16)

    convert_model = auto_convert(convert_model)
    return convert_model



import subprocess

def get_num_nodes():
    return int(subprocess.check_output('lscpu | grep Socket | awk \'{print $2}\'', shell=True))

def get_num_cores_per_node():
    return int(subprocess.check_output('lscpu | grep Core\'.* per socket\' | awk \'{print $4}\'', shell=True))

def get_core_list_of_node_id(node_id):
    r"""
    Helper function to get the CPU cores' ids of the input numa node.

    Args:
        node_id (int): Input numa node id.

    Returns:
        list: List of CPU cores' ids on this numa node.
    """

    num_of_nodes = get_num_nodes()
    assert node_id < num_of_nodes, "input node_id:{0} must less than system number of nodes:{1}".format(node_id, num_of_nodes)
    num_cores_per_node = get_num_cores_per_node()
    return list(range(num_cores_per_node * node_id, num_cores_per_node * (node_id + 1)))



import torch
import intel_extension_for_pytorch._C as core

VERBOSE_OFF = 0
VERBOSE_ON = 1
VERBOSE_ON_CREATION = 2
class verbose(object):
    """
    On-demand oneDNN verbosing functionality

    To make it easier to debug performance issues, oneDNN can dump verbose
    messages containing information like kernel size, input data size and
    execution duration while executing the kernel. The verbosing functionality
    can be invoked via an environment variable named `DNNL_VERBOSE`. However,
    this methodology dumps messages in all steps. Those are a large amount of
    verbose messages. Moreover, for investigating the performance issues,
    generally taking verbose messages for one single iteration is enough.

    This on-demand verbosing functionality makes it possible to control scope
    for verbose message dumping. In the following example, verbose messages
    will be dumped out for the second inference only.

    .. highlight:: python
    .. code-block:: python

        import intel_extension_for_pytorch as ipex
        model(data)
        with ipex.verbose(ipex.VERBOSE_ON):
            model(data)

    Args:
        level: Verbose level

            - ``VERBOSE_OFF``: Disable verbosing
            - ``VERBOSE_ON``:  Enable verbosing
            - ``VERBOSE_ON_CREATION``: Enable verbosing, including oneDNN kernel creation

    :meta public:
    """
    def __init__(self, level):
        self.level = level

    def __enter__(self):
        if self.level == VERBOSE_OFF:
            return
        try:
            st = torch._C._verbose.mkldnn_set_verbose(self.level)
            assert bool(st), "Failed to set Verbose mode of MKLDNN in PyTorch. Please consider to disable this verbose scope."
        except:
            pass
        st = core.mkldnn_set_verbose(self.level)
        assert bool(st), "Failed to set Verbose mode of MKLDNN in IPEX. Please consider to disable this verbose scope."
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        core.mkldnn_set_verbose(VERBOSE_OFF)
        try:
            torch._C._verbose.mkldnn_set_verbose(VERBOSE_OFF)
        except:
            pass
        return False

try:
    verbose_torch = torch.backends.mkldnn.verbose
    torch.backends.mkldnn.verbose = verbose
except:
    pass




import torch
from torch import Tensor, nn
from torch.autograd import Function
from typing import List, Optional, NamedTuple
from itertools import accumulate
import enum

class PoolingMode(enum.IntEnum):
    SUM = 0
    MEAN = 1

class SGDArgs(NamedTuple):
    bf16_trail: List[Optional[torch.Tensor]]
    weight_decay: float
    lr: float

class EmbeddingSpec(NamedTuple):
    num_of_features: int
    feature_size: int
    pooling_modes: str
    dtype: torch.dtype
    weight: Optional[torch.Tensor]

def merged_embeddingbag_sgd(
    indices,
    offsets,
    indices_with_row_offsets,
    row_offsets,
    pooling_modes,
    sgd_args,
    *weights
):
    if torch.is_grad_enabled():
        return MergedEmbeddingBagSGDFunc.apply(
            indices, offsets, indices_with_row_offsets, row_offsets, pooling_modes, sgd_args, *weights
        )
    return torch.ops.torch_ipex.merged_embeddingbag_forward(indices, offsets, weights, pooling_modes)

class MergedEmbeddingBagSGDFunc(Function):
    @staticmethod
    def unpack(*args):
        return args

    @staticmethod
    def forward(ctx, indices, offsets, indices_with_row_offsets, row_offsets, pooling_modes, sgd_args, *weights):
        output = torch.ops.torch_ipex.merged_embeddingbag_forward(
            indices, offsets, weights, pooling_modes
        )
        ctx.indices = indices
        ctx.offsets = offsets
        ctx.weights = weights
        ctx.indices_with_row_offsets = indices_with_row_offsets
        ctx.row_offsets = row_offsets
        ctx.pooling_modes = pooling_modes
        ctx.sgd_args = sgd_args
        return MergedEmbeddingBagSGDFunc.unpack(*output)

    @staticmethod
    def backward(ctx, *grad_out):
        indices = ctx.indices
        offsets = ctx.offsets
        weights = ctx.weights
        indices_with_row_offsets = ctx.indices_with_row_offsets
        row_offsets = ctx.row_offsets
        pooling_modes = ctx.pooling_modes
        sgd_args = ctx.sgd_args
        bf16_trail = sgd_args.bf16_trail
        weight_decay = sgd_args.weight_decay
        lr = sgd_args.lr
        torch.ops.torch_ipex.merged_embeddingbag_backward_sgd(
            grad_out, indices, offsets, weights, indices_with_row_offsets,
            row_offsets, pooling_modes,
            bf16_trail, weight_decay, lr)
        n_tables = len(weights)
        output = [None for i in range(n_tables + 6)]
        return MergedEmbeddingBagSGDFunc.unpack(*output)

class MergedEmbeddingBag(nn.Module):
    r"""
    Merge multiple Pytorch EmbeddingBag (https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/sparse.py#L221) 
    as one torch.nn.Module. 
    Native usage for multiple EmbeddingBag is:
        >>> EmbLists = torch.nn.Modulist(emb1, emb2, emb3, ..., emb_m)
        >>> inputs = [in1, in2, in3, ..., in_m]
        >>> outputs = []
        >>> for i in range(len(EmbLists)):
        >>>     outputs.append(Emb[in_i])
    Our optimized path will be:
        >>> merged_emb = MergedEmbeddingBagWithSGD(args)
        >>> outputs = MergedEmbeddingBagWithSGD(input)
    We will have benefits for our optimized path:
      1). We will save Pytorch OP dispatch overhead, if the EmbeddingBag operations are not
      heavy, this dispatch overhead will have big impact.
    We introduce "linearize_indices_and_offsets" step to merged indices/offsets together. But consider EmbeddingBags
    are usually the first layer of a model. So "linearize_indices_and_offsets" can be considered as "data prepocess" and
    can be done offline.
    This Module can not be used alone, we suggest to use MergedEmbeddingBagWith[Optimizer] instead.
    Now we can only choose MergedEmbeddingBagWithSGD and we plan to add more optimizer support
    in the future.
    For the introduction of MergedEmbeddingBagWith[Optimizer], please find the comments at
    MergedEmbeddingBagWithSGD.
    """
    embedding_specs: List[EmbeddingSpec]

    def __init__(
        self,
        embedding_specs: List[EmbeddingSpec],
    ):
        super(MergedEmbeddingBag, self).__init__()
        self.n_tables = len(embedding_specs)
        self.weights = []
        row_offsets = []
        feature_sizes = []
        self.pooling_modes = []
        self.dtypes = []
        dtype = None
        self.weights = torch.nn.ParameterList([nn.Parameter(torch.Tensor()) for i in range(len(embedding_specs))])
        for i, emb in enumerate(embedding_specs):
            num_of_features, feature_size, mode, dtype, weight = emb
            row_offsets.append(num_of_features)
            if mode == 'sum':
                self.pooling_modes.append(PoolingMode.SUM)
            elif mode == 'mean':
                self.pooling_modes.append(PoolingMode.MEAN)
            else:
                assert False, r"MergedEmbeddingBag only support EmbeddingBag with model sum or mean"
            if weight is None:
                weight = torch.empty((num_of_features, feature_size), dtype=dtype)
            self.weights[i] = nn.Parameter(weight)
        self.register_buffer(
            "row_offsets",
            torch.tensor([0] + list(accumulate(row_offsets)), dtype=torch.int64),
        )

    def extra_repr(self) -> str:
        s = 'number of tables={}\n'.format(self.n_tables)
        for i in range(self.n_tables):
            s += "table{}: {}, {}, {}, {}".format(
                i, self.weights[i].shape[0], self.weights[i].shape[1], self.pooling_modes[i], self.weights[i].dtype)
            if i != self.n_tables - 1:
                s += '\n'
        return s

    def linearize_indices_and_offsets(
        self,
        indices: List[Tensor],
        offsets: List[Optional[Tensor]],
        include_last_offsets: List[bool]
    ):
        r"""
        To make backward/update more balance, we only have 1 logical table in MergedEmbedingBag and
        use unified indices for access the whole logical table.
        We need to re-mark the indice from different tables to distinguish them.
        For example, we have  2 tables with shape [200, 128] and [100, 128].
        The indice 50 for table1 is still 50 and the indice 50 for table2 should be set to 50 + 200 = 250.
        We assume the original indice and offset will follow the usage for Pytorch EmbeddingBag:
        https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/sparse.py#L355-L382
        """
        # TODO: support per_sample_weights in forward
        def get_batch_size(indice, offset, include_last_offset):
            if indice.dim() == 2:
                assert offset is None, "offset should be None if indice is 2-D tensor, https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/sparse.py#L355-L382"
                batch_size = indice.shape[0]
            else:
                batch_size = offset.numel()
                if include_last_offset:
                    batch_size -= 1
            return batch_size

        assert self.n_tables == len(indices), "expected {} but got {} indices".format(self.n_tables, len(indices))
        assert self.n_tables == len(offsets), "expected {} but got {} offsets".format(self.n_tables, len(offsets))
        assert self.n_tables == len(include_last_offsets), "expected {} but got {} include_last_offsets".format(
            self.n_tables, len(include_last_offsets))

        batch_size = get_batch_size(indices[0], offsets[0], include_last_offsets[0])
        assert all(
            batch_size == get_batch_size(idx, offset, include_last) for idx, offset, include_last in zip(indices, offsets, include_last_offsets)
        ), r"MergedEmbeddingBag only support input with same batch size"
        n_indices = sum([t.numel() for t in indices])
        n_offsets = batch_size * self.n_tables + 1  # include last offset
        merged_indices = torch.empty(n_indices, dtype=torch.int64)
        merged_indices_with_row_offsets = torch.empty(n_indices, dtype=torch.int64)  # used for sort together
        merged_offsets = torch.empty(n_offsets, dtype=torch.int64)
        idx_start = 0
        offset_start = 0
        for i in range(self.n_tables):
            n_indice = indices[i].numel()
            merged_indices[idx_start: idx_start + n_indice].copy_(indices[i].view(-1))
            merged_indices_with_row_offsets[idx_start: idx_start + n_indice].copy_(indices[i].view(-1) + self.row_offsets[i])
            if indices[i].dim() == 2:
                bag_size = indices[i].shape[1]
                offset = torch.arange(0, indices[i].numel(), bag_size)
            else:
                offset = offsets[i][:-1] if include_last_offsets[i] else offsets[i]
            assert offset.numel() == batch_size
            merged_offsets[offset_start : offset_start + batch_size].copy_(offset + idx_start)
            idx_start += n_indice
            offset_start += batch_size
        assert idx_start == n_indices
        assert offset_start == n_offsets - 1
        merged_offsets[-1] = n_indices
        return (merged_indices, merged_offsets, merged_indices_with_row_offsets)

    def forward(self, input, need_linearize_indices_and_offsets=torch.BoolTensor([True])):
        assert False, "Please use MergedEmbeddingBagWith[Optimizer]. We only support SGD now, so please create module MergedEmbeddingBagWithSGD instead"


class MergedEmbeddingBagWithSGD(MergedEmbeddingBag):
    r"""
    To support training for MergedEmbeddingBag with good performance, we fused optimizer step
    with backward function.
    Native usage for multiple EmbeddingBag is:
        >>> EmbLists = torch.nn.Modulist(emb1, emb2, emb3, ..., emb_m)
        >>> sgd = torch.optim.SGD(EmbLists.parameters(), lr=lr, weight_decay=weight_decay)
        >>> inputs = [in1, in2, in3, ..., in_m]
        >>> outputs = []
        >>> for i in range(len(EmbLists)):
        >>>     outputs.append(Emb[in_i])
        >>> sgd.zero_grad()
        >>> for i in range(len(outputs)):
        >>>     out.backward(grads[i]) 
        >>> sgd.step()
    Our optimized path will be:
        >>> # create MergedEmbeddingBagWithSGD module with optimizer args (lr and weight decay)
        >>> merged_emb = MergedEmbeddingBagWithSGD(args)
        >>> merged_input = merged_emb.linearize_indices_and_offsets(inputs)
        >>> outputs = MergedEmbeddingBagWithSGD(merged_input)
        >>> outputs.backward(grads)
    We will get further benefits in training:
      1). We will futher save Pytorch OP dispatch overhead in backward and weight update process.
      2). We will make thread loading more balance during backward/weight update. In real
      world scenario, Embedingbag are often used to represent categorical features and the 
      categorical features will often fit power law distribution. For example, if we use one
      Embeddingtable to represent the age range the users of a video game website. We might
      find most of users are from 10-19 or 20-29. So we may need update the row which represent
      10-19 or 20-29 frequently. Since update these rows need to write at the same memory address,
      we need to write it by 1 thread (or we will have write conflict or have overhead to solve the conflict).
      By merge multiple table together, we will have more friendly distribution to distribute
      backward/update tasks.
      3). We will fuse update with backward together. We can immediately update the weight after
      we get grad from backward thus the memory pattern will be more friendly. We will have 
      more chance to access data from cache. 
    """
    embedding_specs: List[EmbeddingSpec]

    def __init__(
        self,
        embedding_specs: List[EmbeddingSpec],
        lr: float = 0.01,
        weight_decay: float = 0
    ):
        super(MergedEmbeddingBagWithSGD, self).__init__(embedding_specs)
        self.sgd_args = self.init_sgd_args(lr, weight_decay)
        for i in range(self.n_tables):
            weight = self.weights[i]
            if weight.dtype == torch.bfloat16:
                self.sgd_args.bf16_trail.append(torch.zeros_like(weight, dtype=torch.bfloat16))
            else:
                self.sgd_args.bf16_trail.append(torch.empty(0, dtype=torch.bfloat16))

    def init_sgd_args(self, lr, weight_decay, bf16_trail=[]):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        return SGDArgs(
            weight_decay=weight_decay,
            lr=lr,
            bf16_trail=bf16_trail
        )

    def to_bfloat16_train(self):
        r"""
        Cast weight to bf16 and it's trail part for training
        """
        trails = []
        for i in range(len(self.weights)):
            if self.weights[i].dtype == torch.float:
                bf16_w, trail = torch.ops.torch_ipex.split_float_bfloat16(self.weights[i])
            elif self.weights[i].dtype == torch.bfloat16:
                bf16_w = self.weights[i]
                trail = torch.zeros_like(bf16_w, dtype=torch.bfloat16)
            elif self.weights[i].dtype == torch.double:
                bf16_w, trail = torch.ops.torch_ipex.split_float_bfloat16(self.weights[i].float())
            else:
                assert False, r"MergedEmbeddingBag only support dtypes with bfloat, float and double"
            trails.append(trail)
            self.weights[i] = torch.nn.Parameter(bf16_w)
        self.sgd_args = self.sgd_args._replace(bf16_trail=trails)

    def forward(self, input, need_linearize_indices_and_offsets=torch.BoolTensor([True])):
        r"""
        Args:
            input (Tuple[Tensor]): a tuple of (indices, offsets, include_last_offsets(if not merged)/indices_with_row_offsets(if merged))
            need_linearize_indices_and_offsets: indicate whether input need to be linearized
        Returns:
            List[Tensor] output shape of `(batch_size, feature_size)` which length = num of tables.
        """
        if need_linearize_indices_and_offsets.item():
            indices, offsets, include_last_offsets = input
            indices, offsets, indices_with_row_offsets = self.linearize_indices_and_offsets(indices, offsets, include_last_offsets)
        else:
            indices, offsets, indices_with_row_offsets = input
        return merged_embeddingbag_sgd(
            indices, offsets, indices_with_row_offsets, self.row_offsets,
            self.pooling_modes, self.sgd_args, *self.weights
        )

    @classmethod
    def from_embeddingbag_list(
        cls,
        tables: List[torch.nn.EmbeddingBag],
        lr: float = 0.01,
        weight_decay: float = 0
    ):
        embedding_specs = []
        for emb in tables:
            emb_shape = emb.weight.shape
            embedding_specs.append(
                EmbeddingSpec(
                    num_of_features=emb_shape[0],
                    feature_size=emb_shape[1],
                    pooling_modes=emb.mode,
                    dtype=emb.weight.dtype,
                    weight=emb.weight.detach()
                ))
        return cls(embedding_specs, lr, weight_decay)



import torch
import intel_extension_for_pytorch._C as core

class _profiler(torch.profiler.profile):
    def __enter__(self, *args):
        if(self.activities):
            for i in self.activities:
                if(i is torch.profiler.ProfilerActivity.CPU):
                    core.set_profile_op_enabled(True)
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()
        if(self.activities):
            for i in self.activities:
                if(i is torch.profiler.ProfilerActivity.CPU):
                    core.set_profile_op_enabled(False)

torch.profiler.profile = _profiler



import torch
import warnings

def _numpy(x):
    if x.dtype==torch.bfloat16:
        warnings.warn("calling in ipex numpy which is not share mermory with torch tensor for bfloat16 input.")
        return torch._C._TensorBase.numpy(x.float())
    else:
        return torch._C._TensorBase.numpy(x)

torch.Tensor.numpy = _numpy




from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import platform
import subprocess
import os
from os.path import expanduser
import re
import glob
import numpy as np
from argparse import ArgumentParser, REMAINDER
from argparse import RawTextHelpFormatter
import logging
import psutil
from datetime import datetime

format_str = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(level=logging.INFO, format=format_str)
logger = logging.getLogger(__name__)

r"""
This is a script for launching PyTorch training and inference on Intel Xeon CPU with optimal configurations.
Now, single instance inference/training, multi-instance inference/training and distributed training
with oneCCL backend is enabled.

To get the peak performance on Intel Xeon CPU, the script optimizes the configuration of thread and memory
management. For thread management, the script configures thread affinity and the preload of Intel OMP library.
For memory management, it configures NUMA binding and preload optimized memory allocation library (e.g. tcmalloc, jemalloc).

**How to use this module:**

*** Single instance inference/training ***

1. Run single-instance inference or training on a single node with all CPU nodes.

::

   >>> python -m intel_extension_for_pytorch.cpu.launch --throughput_mode script.py args

2. Run single-instance inference or training on a single CPU node.

::

   >>> python -m intel_extension_for_pytorch.cpu.launch --node_id 1 script.py args

*** Multi-instance inference ***

1. Multi-instance
   By default, one instance per node. if you want to set the instance numbers and core per instance,
   --ninstances and  --ncore_per_instance should be set.


   >>> python -m intel_extension_for_pytorch.cpu.launch -- python_script args

   eg: on CLX8280 with 14 instance, 4 cores per instance
::

   >>> python -m intel_extension_for_pytorch.cpu.launch  --ninstances 14 --ncore_per_instance 4 python_script args

2. Run single-instance inference among multiple instances.
   By default, runs all ninstances. If you want to independently run a single instance among ninstances, specify instance_idx.

   eg: run 0th instance among SKX with 2 instance (i.e., numactl -C 0-27)
::

   >>> python -m intel_extension_for_pytorch.cpu.launch  --ninstances 2 --instance_idx 0 python_script args

   eg: run 1st instance among SKX with 2 instance (i.e., numactl -C 28-55)
::

   >>> python -m intel_extension_for_pytorch.cpu.launch  --ninstances 2 --instance_idx 1 python_script args

   eg: run 0th instance among SKX with 2 instance, 2 cores per instance, first four cores (i.e., numactl -C 0-1)
::

   >>> python -m intel_extension_for_pytorch.cpu.launch  --core_list "0, 1, 2, 3" --ninstances 2 --ncore_per_instance 2 --instance_idx 0 python_script args

*** Distributed Training ***

spawns up multiple distributed training processes on each of the training nodes. For intel_extension_for_pytorch, oneCCL
is used as the communication backend and MPI used to launch multi-proc. To get the better
performance, you should specify the different cores for oneCCL communication and computation
process seperately. This tool can automatically set these ENVs(such as I_MPI_PIN_DOMIN) and launch
multi-proc for you.

The utility can be used for single-node distributed training, in which one or
more processes per node will be spawned.  It can also be used in
multi-node distributed training, by spawning up multiple processes on each node
for well-improved multi-node distributed training performance as well.


1. Single-Node multi-process distributed training

::

    >>> python  -m intel_extension_for_pytorch.cpu.launch --distributed  python_script  --arg1 --arg2 --arg3 and all other
                arguments of your training script

2. Multi-Node multi-process distributed training: (e.g. two nodes)


rank 0: *(IP: 192.168.10.10, and has a free port: 295000)*

::

    >>> python -m intel_extension_for_pytorch.cpu.launch --distributed --nproc_per_node=xxx
               --nnodes=2 --hostfile hostfile python_sript --arg1 --arg2 --arg3
               and all other arguments of your training script)


3. To look up what optional arguments this module offers:

::

    >>> python -m intel_extension_for_pytorch.cpu.launch --help

*** Memory allocator  ***

"--enable_tcmalloc" and "--enable_jemalloc" can be used to enable different memory allcator.

"""

class CPUinfo():
    '''
    Get CPU inforamation, such as cores list and NUMA information.
    '''
    def __init__(self):

        self.cpuinfo = []
        if platform.system() == "Windows":
            raise RuntimeError("Windows platform is not supported!!!")
        elif platform.system() == "Linux":
            args = ["lscpu", "--parse=CPU,Core,Socket,Node"]
            lscpu_info = subprocess.check_output(args, universal_newlines=True).split("\n")

            # Get information about  cpu, core, socket and node
            for line in lscpu_info:
                pattern = r"^([\d]+,[\d]+,[\d]+,[\d]?)"
                regex_out = re.search(pattern, line)
                if regex_out:
                    self.cpuinfo.append(regex_out.group(1).strip().split(","))
            assert len(self.cpuinfo) > 0, "cpuinfo is empty"
            self.get_socket_info()

    def get_socket_info(self):
        idx_active = 3
        if self.cpuinfo[0][idx_active] == '':
            idx_active = 2
        self.nodes = int(max([line[idx_active] for line in self.cpuinfo])) + 1
        self.node_physical_cores = []  # node_id is index
        self.node_logical_cores = []   # node_id is index
        self.physical_core_node_map = {}  # phyical core to numa node id
        self.logical_core_node_map = {}   # logical core to numa node id

        for node_id in range(self.nodes):
            cur_node_physical_core = []
            cur_node_logical_core = []
            for line in self.cpuinfo:
                nid = line[idx_active] if line[idx_active] != '' else '0'
                if node_id == int(nid):
                    if int(line[1]) not in cur_node_physical_core:
                        cur_node_physical_core.append(int(line[1]))
                        self.physical_core_node_map[int(line[1])] = int(node_id)
                    cur_node_logical_core.append(int(line[0]))
                    self.logical_core_node_map[int(line[0])] = int(node_id)
            self.node_physical_cores.append(cur_node_physical_core)
            self.node_logical_cores.append(cur_node_logical_core)

    def node_nums(self):
        return self.nodes

    def physical_core_nums(self):
        return len(self.node_physical_cores) * len(self.node_physical_cores[0])

    def logical_core_nums(self):
        return len(self.node_logical_cores) * len(self.node_logical_cores[0])

    def get_node_physical_cores(self, node_id):
        if node_id < 0 or node_id > self.nodes - 1:
            logger.error("Invalid node id")
        return self.node_physical_cores[node_id]

    def get_node_logical_cores(self, node_id):
        if node_id < 0 or node_id > self.nodes - 1:
            logger.error("Invalid node id")
        return self.node_logical_cores[node_id]

    def get_all_physical_cores(self):
        return np.array(self.node_physical_cores).flatten().tolist()

    def get_all_logical_cores(self):
        return np.array(self.node_logical_cores).flatten().tolist()

    def numa_aware_check(self, core_list):
        '''
        Check whether all cores in core_list are in the same NUMA node. cross NUMA will reduce perforamnce.
        We strongly advice to not use cores on different nodes.
        '''
        cores_numa_map = self.logical_core_node_map
        numa_ids = []
        for core in core_list:
            numa_id = cores_numa_map[core]
            if not numa_id in numa_ids:
                numa_ids.append(numa_id)
        if len(numa_ids) > 1:
            logger.warning("Numa Aware: cores:{} on different NUMA nodes:{}".format(str(core_list), str(numa_ids)))
        if len(numa_ids) == 0:
            logger.error("invalid number of NUMA nodes; please make sure numa_ids >= 1")
            exit(-1)
        return numa_ids

class Launcher():
    r"""
     Base class for launcher
    """
    def __init__(self):
        self.cpuinfo = CPUinfo()

    def launch(self, args):
        pass

    def add_lib_preload(self, lib_type=None):
        '''
        Enale TCMalloc/JeMalloc/intel OpenMP
        '''
        library_paths = []
        if "CONDA_PREFIX" in os.environ:
            library_paths.append(os.environ["CONDA_PREFIX"] + "/lib/")
        if "VIRTUAL_ENV" in os.environ:
            library_paths.append(os.environ["VIRTUAL_ENV"] + "/lib/")

        library_paths += ["{}/.local/lib/".format(expanduser("~")), "/usr/local/lib/",
                          "/usr/local/lib64/", "/usr/lib/", "/usr/lib64/"]

        lib_find = False
        lib_set = False
        for item in os.getenv("LD_PRELOAD", "").split(":"):
            if item.endswith('lib{}.so'.format(lib_type)):
                lib_set = True
                break
        if not lib_set:
            for lib_path in library_paths:
                library_file = lib_path + "lib" + lib_type + ".so"
                matches = glob.glob(library_file)
                if len(matches) > 0:
                    if "LD_PRELOAD" in os.environ:
                        os.environ["LD_PRELOAD"] = matches[0] + ":" + os.environ["LD_PRELOAD"]
                    else:
                        os.environ["LD_PRELOAD"] = matches[0]
                    lib_find = True
                    break
        return lib_set or lib_find

    def is_numactl_available(self):
        numactl_available = False
        cmd = ["numactl", "-C", "0", "-m", "0", "ls"]
        r = subprocess.run(cmd, env=os.environ, stdout=subprocess.DEVNULL)
        if r.returncode == 0:
            numactl_available = True
        return numactl_available

    def set_memory_allocator(self, enable_tcmalloc=True, enable_jemalloc=False, use_default_allocator=False):
        '''
        Enable TCMalloc/JeMalloc with LD_PRELOAD and set configuration for JeMalloc.
        By default, PTMalloc will be used for PyTorch, but TCMalloc and JeMalloc can get better
        memory resue and reduce page fault to improve performance.
        '''
        if enable_tcmalloc and enable_jemalloc:
            logger.error("Unable to enable TCMalloc and JEMalloc at the same time")
            exit(-1)

        if enable_tcmalloc:
            find_tc = self.add_lib_preload(lib_type="tcmalloc")
            if not find_tc:
                logger.warning("Unable to find the {} library file lib{}.so in $CONDA_PREFIX/lib or $VIRTUAL_ENV/lib"
                               " or /.local/lib/ or /usr/local/lib/ or /usr/local/lib64/ or /usr/lib or /usr/lib64 or "
                               "{}/.local/lib/ so the LD_PRELOAD environment variable will not be set."
                               "you can use 'conda install -c conda-forge gperftools' to install tcmalloc"
                               .format("TCmalloc", "tcmalloc", expanduser("~")))
            else:
                logger.info("Use TCMalloc memory allocator")

        elif enable_jemalloc:
            find_je = self.add_lib_preload(lib_type="jemalloc")
            if not find_je:
                logger.warning("Unable to find the {} library file lib{}.so in $CONDA_PREFIX/lib or $VIRTUAL_ENV/lib"
                               " or /.local/lib/ or /usr/local/lib/ or /usr/local/lib64/ or /usr/lib or /usr/lib64 or "
                               "{}/.local/lib/ so the LD_PRELOAD environment variable will not be set."
                               "you can use 'conda install -c conda-forge jemalloc' to install jemalloc"
                               .format("JeMalloc", "jemalloc", expanduser("~")))
            else:
                logger.info("Use JeMalloc memory allocator")
                self.set_env('MALLOC_CONF', "oversize_threshold:1,background_thread:true,metadata_thp:auto")

        elif use_default_allocator:
            pass

        else:
            find_tc = self.add_lib_preload(lib_type="tcmalloc")
            if find_tc:
                logger.info("Use TCMalloc memory allocator")
                return
            find_je = self.add_lib_preload(lib_type="jemalloc")
            if find_je:
                logger.info("Use JeMalloc memory allocator")
                return
            logger.warning("Neither TCMalloc nor JeMalloc is found in $CONDA_PREFIX/lib or $VIRTUAL_ENV/lib"
                           " or /.local/lib/ or /usr/local/lib/ or /usr/local/lib64/ or /usr/lib or /usr/lib64 or "
                           "{}/.local/lib/ so the LD_PRELOAD environment variable will not be set. This may drop the performance"
                           .format(expanduser("~")))

    def logger_env(self, env_name=""):
        if env_name in os.environ:
            logger.info("{}={}".format(env_name, os.environ[env_name]))

    def set_env(self, env_name, env_value=None):
        if not env_value:
            logger.warning("{} is None".format(env_name))
        if env_name not in os.environ:
            os.environ[env_name] = env_value
        elif os.environ[env_name] != env_value:
            logger.warning("{} in environment variable is {} while the value you set is {}".format(env_name, os.environ[env_name], env_value))
        self.logger_env(env_name)

    # set_kmp_affinity is used to control whether to set KMP_AFFINITY or not. In scenario that use all cores on all nodes, including logical cores, setting KMP_AFFINITY disables logical cores. In this case, KMP_AFFINITY should not be set.
    def set_multi_thread_and_allocator(self, ncore_per_instance, disable_iomp=False, set_kmp_affinity=True, enable_tcmalloc=True, enable_jemalloc=False, use_default_allocator=False):
        '''
        Set multi-thread configuration and enable Intel openMP and TCMalloc/JeMalloc.
        By default, GNU openMP and PTMalloc are used in PyTorch. but Intel openMP and TCMalloc/JeMalloc are better alternatives
        to get performance benifit.
        '''
        self.set_memory_allocator(enable_tcmalloc, enable_jemalloc, use_default_allocator)
        self.set_env("OMP_NUM_THREADS", str(ncore_per_instance))
        if not disable_iomp:
            find_iomp = self.add_lib_preload(lib_type="iomp5")
            if not find_iomp:
                logger.warning("Unable to find the {} library file lib{}.so in $CONDA_PREFIX/lib or $VIRTUAL_ENV/lib"
                               " or /.local/lib/ or /usr/local/lib/ or /usr/local/lib64/ or /usr/lib or /usr/lib64 or "
                               "{}/.local/lib/ so the LD_PRELOAD environment variable will not be set."
                               "you can use 'conda install intel-openm' to install intel openMP"
                               .format("iomp", "iomp5", expanduser("~")))
            else:
                logger.info("Using Intel OpenMP")
                if set_kmp_affinity:
                    self.set_env("KMP_AFFINITY", "granularity=fine,compact,1,0")
                self.set_env("KMP_BLOCKTIME", "1")
        self.logger_env("LD_PRELOAD")

class MultiInstanceLauncher(Launcher):
    r"""
     Launcher for single instance and multi-instance
     """
    def launch(self, args):
        processes = []
        cores = []
        set_kmp_affinity = True
        enable_taskset = False
        if args.core_list:  # user specify what cores will be used by params
            cores = [int(x) for x in args.core_list.split(",")]
            if args.ncore_per_instance == -1:
                logger.error("please specify the '--ncore_per_instance' if you have pass the --core_list params")
                exit(-1)
            elif args.ninstances > 1 and args.ncore_per_instance * args.ninstances < len(cores):
                logger.warning("only first {} cores will be used, but you specify {} cores in core_list".format(args.ncore_per_instance * args.ninstances, len(cores)))
            else:
                args.ninstances = len(cores) // args.ncore_per_instance

        else:
            if args.use_logical_core:
                if args.node_id != -1:
                    cores = self.cpuinfo.get_node_logical_cores(args.node_id)
                else:
                    cores = self.cpuinfo.get_all_logical_cores()
                    # When using all cores on all nodes, including logical cores, setting KMP_AFFINITY disables logical cores. Thus, KMP_AFFINITY should not be set.
                    set_kmp_affinity = False
            else:
                if args.node_id != -1:
                    cores = self.cpuinfo.get_node_physical_cores(args.node_id)
                else:
                    cores = self.cpuinfo.get_all_physical_cores()
            if not args.multi_instance and args.ninstances == -1 and args.ncore_per_instance == -1:
                args.ninstances = 1
                args.ncore_per_instance = len(cores)
            elif args.multi_instance and args.ninstances == -1 and args.ncore_per_instance == -1:
                args.throughput_mode = True
            elif args.ncore_per_instance == -1 and args.ninstances != -1:
                if args.ninstances > len(cores):
                    logger.error("there are {} total cores but you specify {} ninstances; please make sure ninstances <= total_cores)".format(len(cores), args.ninstances))
                    exit(-1)
                else:
                    args.ncore_per_instance = len(cores) // args.ninstances
            elif args.ncore_per_instance != -1 and args.ninstances == -1:
                if not args.skip_cross_node_cores:
                    args.ninstances = len(cores) // args.ncore_per_instance
                else:
                    ncore_per_node = len(self.cpuinfo.node_physical_cores[0])
                    num_leftover_cores = ncore_per_node % args.ncore_per_instance
                    if args.ncore_per_instance > ncore_per_node:
                        # too many ncore_per_instance to skip cross-node cores
                        logger.warning("there are {} core(s) per socket, but you specify {} ncore_per_instance and skip_cross_node_cores. Please make sure --ncore_per_instance < core(s) per socket".format(ncore_per_node, args.ncore_per_instance))
                        exit(-1)
                    elif num_leftover_cores == 0:
                        # aren't any cross-node cores
                        logger.info('--skip_cross_node_cores is set, but there are no cross-node cores.')
                        args.ninstances = len(cores) // args.ncore_per_instance
                    else:
                        # skip cross-node cores
                        if args.ninstances != -1:
                            logger.warning('--skip_cross_node_cores is exclusive to --ninstances. --ninstances won\'t take effect even if it is set explicitly.')

                        i = 1
                        leftover_cores = set()
                        while ncore_per_node*i <= len(cores):
                            leftover_cores.update(cores[ncore_per_node*i-num_leftover_cores : ncore_per_node*i])
                            i += 1
                        cores = list(set(cores) - leftover_cores)
                        assert len(cores) % args.ncore_per_instance == 0
                        args.ninstances = len(cores) // args.ncore_per_instance
            else:
                if args.ninstances * args.ncore_per_instance > len(cores):
                    logger.error("Please make sure ninstances * ncore_per_instance <= total_cores")
                    exit(-1)
            if args.latency_mode:
                logger.warning('--latency_mode is exclusive to --ninstances, --ncore_per_instance, --node_id and --use_logical_core. They won\'t take effect even if they are set explicitly.')
                args.ncore_per_instance = 4
                cores = self.cpuinfo.get_all_physical_cores()
                args.ninstances = len(cores) // args.ncore_per_instance

            if args.throughput_mode:
                logger.warning('--throughput_mode is exclusive to --ninstances, --ncore_per_instance, --node_id and --use_logical_core. They won\'t take effect even if they are set explicitly.')
                args.ninstances = self.cpuinfo.node_nums()
                cores = self.cpuinfo.get_all_physical_cores()
                args.ncore_per_instance = len(cores) // args.ninstances

        if args.ninstances > 1 and args.instance_idx != -1:
            logger.info("assigning {} cores for instance {}".format(args.ncore_per_instance, args.instance_idx))

        if not args.disable_numactl:
            numactl_available = self.is_numactl_available()
            if not numactl_available:
                if not args.disable_taskset:
                    logger.warning("Core binding with numactl is not available. Disabling numactl and using taskset instead. This may affect performance in multi-socket system; please use numactl if memory binding is needed.")
                    args.disable_numactl = True
                    enable_taskset = True
                else:
                    logger.warning("Core binding with numactl is not available, and --disable_taskset is set. Please unset --disable_taskset to use taskset insetad of numactl.")
                    exit(-1)

        if not args.disable_taskset:
            enable_taskset = True

        self.set_multi_thread_and_allocator(args.ncore_per_instance,
                                            args.disable_iomp,
                                            set_kmp_affinity,
                                            args.enable_tcmalloc,
                                            args.enable_jemalloc,
                                            args.use_default_allocator)
        os.environ["LAUNCH_CMD"] = "#"
        for i in range(args.ninstances):
            cmd = []
            cur_process_cores = ""
            if not args.disable_numactl or enable_taskset:
                if not args.disable_numactl:
                    cmd = ["numactl"]
                elif enable_taskset:
                    cmd = ["taskset"]

                cores = sorted(cores)
                if args.instance_idx == -1:  # sequentially assign ncores_per_instance to ninstances
                    core_list = cores[i * args.ncore_per_instance: (
                        i + 1) * args.ncore_per_instance]
                else:  # assign ncores_per_instance from instance_idx
                    core_list = cores[args.instance_idx * args.ncore_per_instance: (
                        args.instance_idx + 1) * args.ncore_per_instance]

                core_ranges = []
                for core in core_list:
                    if len(core_ranges) == 0:
                        range_elem = {'start': core, 'end': core}
                        core_ranges.append(range_elem)
                    else:
                        if core - core_ranges[-1]['end'] == 1:
                            core_ranges[-1]['end'] = core
                        else:
                            range_elem = {'start': core, 'end': core}
                            core_ranges.append(range_elem)
                for r in core_ranges:
                    cur_process_cores = cur_process_cores + "{}-{},".format(r['start'], r['end'])
                cur_process_cores = cur_process_cores[:-1]

                if not args.disable_numactl:
                    numa_params = "-C {} ".format(cur_process_cores)
                    numa_params += "-m {}".format(",".join(
                        [str(numa_id) for numa_id in self.cpuinfo.numa_aware_check(core_list)]))
                    cmd.extend(numa_params.split())
                elif enable_taskset:
                    taskset_params = "-c {}".format(cur_process_cores)
                    cmd.extend(taskset_params.split())

            with_python = not args.no_python
            if with_python:
                cmd.append(sys.executable)
                cmd.append("-u")
            if args.module:
                cmd.append("-m")
            cmd.append(args.program)
            log_name = args.log_file_prefix + "_instance_{}_cores_".format(i) + cur_process_cores.replace(',', '_') + ".log"
            log_name = os.path.join(args.log_path, log_name)
            cmd.extend(args.program_args)
            os.environ["LAUNCH_CMD"] += " ".join(cmd) + ",#"
            cmd_s = " ".join(cmd)
            if args.log_path:
                cmd_s = "{} 2>&1 | tee {}".format(cmd_s, log_name)
            logger.info(cmd_s)
            if not args.disable_numactl:
                process = subprocess.Popen(cmd_s, env=os.environ, shell=True)
            elif enable_taskset:
                process = subprocess.Popen(cmd, env=os.environ)
            processes.append(process)

            if args.instance_idx != -1: # launches single instance, instance_idx, only
                break

        os.environ["LAUNCH_CMD"] = os.environ["LAUNCH_CMD"][:-2]
        for process in processes:
            process.wait()
            if process.returncode != 0:
                raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd_s)

class DistributedTrainingLauncher(Launcher):
    r"""
     Launcher for distributed traning with MPI launcher
     """
    def get_mpi_pin_domain(self, nproc_per_node, ccl_worker_count, total_cores):
        '''
        I_MPI_PIN_DOMAIN specify the cores used for every MPI process.
        The first ccl_worker_count cores of every rank for ccl communication
        and the other cores will be used to do computation.
        For example: on CascadeLake 8280 CPU, 2 ranks on one node. ccl_worker_count=4
        CCL_WORKER_COUNT=4
        CCL_WORKER_AFFINITY="0,1,2,3,28,29,30,31"
        I_MPI_PIN_DOMAIN=[0xffffff0,0xffffff0000000]
        '''
        ppn = nproc_per_node
        cores_per_rank = total_cores // ppn
        pin_domain = "["
        for proc in range(ppn):
            domain_binary = 0
            begin = proc * cores_per_rank + ccl_worker_count
            end = proc * cores_per_rank + cores_per_rank - 1
            for i in range(begin, end + 1):
                domain_binary |= (1 << i)
            pin_domain += hex(domain_binary) + ","
        pin_domain += "]"
        return pin_domain

    def get_ccl_worker_affinity(self, nproc_per_node, ccl_worker_count, total_cores):
        '''
        Computation and communication use different cores when using oneCCL
        backend for distributed training. we use first ccl_worker_count cores of
        every rank for ccl communication
        '''
        ppn = nproc_per_node
        cores_per_rank = total_cores // ppn
        affinity = ''
        for proc in range(ppn):
            for ccl_worker in range(ccl_worker_count):
                affinity += str(proc * cores_per_rank + ccl_worker) + ","
        affinity = affinity[:-1]
        return affinity

    def launch(self, args):
        '''
        Set ENVs and launch MPI process for distributed training.
        '''
        if args.nnodes > 1 and not os.path.exists(args.hostfile):
            raise ValueError("hostfile is necessary when you use multi-node distributed training,"
                             "Please create hostfile which include the ip list you used for distributed running")
        elif args.nnodes > 1:
            ipv4_addr_pattern = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
            ip_list = []
            with open(args.hostfile) as f:
                for line in f:
                    line = line.strip().strip("\n")
                    # is_valid = re.match(ipv4_addr_pattern, line)
                    # if not is_valid:
                    #     logger.error("{} is not valid IPV4 address".format(line))
                    #     exit(-1)
                    # else:
                    #     ip_list.append(line)
                    ip_list.append(line)
            if len(ip_list) < args.nnodes:
                logger.error("The number of IP {} should greater than nnodes parameters {}".format(len(ip_list), args.nnodes))
                exit(-1)
            master_check = False
            dic = psutil.net_if_addrs()
            for adapter in dic:
                snicList = dic[adapter]
                for snic in snicList:
                    if snic.address == ip_list[0]:
                        master_check = True
            if not master_check:
                logger.error("MASTER_ADDR is incorrect. Please make sure the first line {} in your hostfile is ip address of the current node".format(ip_list[0]))
                exit(-1)

            logger.info("Begin to validate the ip connect")
            args.master_addr = ip_list[0]
            for ip in ip_list[1:]:
                completed_process = subprocess.run("ssh -o PasswordAuthentication=no {} ':'".format(ip), shell=True)
                if completed_process.returncode != 0:
                    logger.error("Passwordless SSH login to {} failed, please make sure you have setup SSH public key right")
                    exit(-1)
                else:
                    logger.info("connection from master node {} to slave node {} is OK".format(args.master_addr, ip))

        total_cores_per_node = self.cpuinfo.physical_core_nums()
        if args.use_logical_core:
            total_cores_per_node = self.cpuinfo.logical_core_nums()

        # set distributed related environmental variables
        self.set_env("MASTER_ADDR", args.master_addr)
        self.set_env("MASTER_PORT", str(args.master_port))
        mpi_pin_domain = self.get_mpi_pin_domain(args.nproc_per_node, args.ccl_worker_count, total_cores_per_node)
        self.set_env("I_MPI_PIN_DOMAIN", mpi_pin_domain)

        ppn = args.nproc_per_node
        cores_per_rank = total_cores_per_node // ppn

        omp_num_threads = cores_per_rank - args.ccl_worker_count
        self.set_multi_thread_and_allocator(omp_num_threads,
                                            args.disable_iomp,
                                            True,
                                            args.enable_tcmalloc,
                                            args.enable_jemalloc,
                                            args.use_default_allocator)

        self.set_env("CCL_WORKER_COUNT", str(args.ccl_worker_count))
        ccl_affinity = self.get_ccl_worker_affinity(args.nproc_per_node, args.ccl_worker_count, total_cores_per_node)
        self.set_env("CCL_WORKER_AFFINITY", ccl_affinity)

        os.environ["LAUNCH_CMD"] = "#"
        cmd = ['mpiexec.hydra']
        mpi_config = "-l -np {} -ppn {} -genv I_MPI_PIN_DOMAIN={} -genv OMP_NUM_THREADS={} ".format(args.nnodes * args.nproc_per_node, args.nproc_per_node, mpi_pin_domain, omp_num_threads)
        mpi_config += args.more_mpi_params
        if args.nnodes > 1:
            mpi_config += " -hostfile {}".format(args.hostfile)
        cmd.extend(mpi_config.split())
        with_python = not args.no_python
        if with_python:
            cmd.append(sys.executable)
            cmd.append("-u")
        if args.module:
            cmd.append("-m")
        cmd.append(args.program)
        cmd.extend(args.program_args)
        log_name = args.log_file_prefix + ".log"
        log_name = os.path.join(args.log_path, log_name)
        cmd_s = " ".join(cmd)
        if args.log_path:
            cmd_s = "{} 2>&1 | tee {}".format(cmd_s, log_name)
        logger.info(cmd_s)
        process = subprocess.Popen(cmd_s, env=os.environ, shell=True)
        process.wait()
        os.environ["LAUNCH_CMD"] += " ".join(cmd) + ",#"
        os.environ["LAUNCH_CMD"] = os.environ["LAUNCH_CMD"][:-2]

def add_distributed_training_params(parser):

    cpuinfo = CPUinfo()
    node_nums = cpuinfo.node_nums()

    group = parser.add_argument_group("Distributed Training Parameters With oneCCL backend")
    group.add_argument("--nnodes", metavar='\b', type=int, default=1,
                       help="The number of nodes to use for distributed "
                       "training")
    group.add_argument("--nproc_per_node", metavar='\b', type=int, default=node_nums,
                       help="The number of processes to launch on each node")
    # ccl control
    group.add_argument("--ccl_worker_count", metavar='\b', default=4, type=int,
                       help="Core numbers per rank used for ccl communication")
    # mpi control
    group.add_argument("--master_addr", metavar='\b', default="127.0.0.1", type=str,
                       help="Master node (rank 0)'s address, should be either "
                            "the IP address or the hostname of node 0, for "
                            "single node multi-proc training, the "
                            "--master_addr can simply be 127.0.0.1")
    group.add_argument("--master_port", metavar='\b', default=29500, type=int,
                       help="Master node (rank 0)'s free port that needs to "
                            "be used for communication during distributed "
                            "training")
    group.add_argument("--hostfile", metavar='\b', default="hostfile", type=str,
                       help="Hostfile is necessary for multi-node multi-proc "
                            "training. hostfile includes the node address list "
                            "node address which should be either the IP address"
                            "or the hostname.")
    group.add_argument("--more_mpi_params", metavar='\b', default="", type=str,
                       help="User can pass more parameters for mpiexec.hydra "
                            "except for -np -ppn -hostfile and -genv I_MPI_PIN_DOMAIN")

def add_memory_allocator_params(parser):

    group = parser.add_argument_group("Memory Allocator Parameters")
    # allocator control
    group.add_argument("--enable_tcmalloc", action='store_true', default=False,
                       help="Enable tcmalloc allocator")
    group.add_argument("--enable_jemalloc", action='store_true', default=False,
                       help="Enable jemalloc allocator")
    group.add_argument("--use_default_allocator", action='store_true', default=False,
                       help="Use default memory allocator")

def add_multi_instance_params(parser):

    group = parser.add_argument_group("Multi-instance Parameters")
    # multi-instance control
    group.add_argument("--ncore_per_instance", metavar='\b', default=-1, type=int,
                       help="Cores per instance")
    group.add_argument("--skip_cross_node_cores", action='store_true', default=False,
                       help="If specified --ncore_per_instance, skips cross-node cores.")
    group.add_argument("--ninstances", metavar='\b', default=-1, type=int,
                       help="For multi-instance, you should give the cores number you used for per instance.")
    group.add_argument("--instance_idx", metavar='\b', default="-1", type=int,
                       help="Specify instance index to assign ncores_per_instance for instance_idx; otherwise ncore_per_instance will be assigned sequentially to ninstances. Please refer to https://github.com/intel/intel-extension-for-pytorch/blob/master/docs/tutorials/performance_tuning/launch_script.md")
    group.add_argument("--latency_mode", action='store_true', default=False,
                       help="By detault 4 core per instance and use all physical cores")
    group.add_argument("--throughput_mode", action='store_true', default=False,
                       help="By default one instance per node and use all physical cores")
    group.add_argument("--node_id", metavar='\b', default=-1, type=int,
                       help="node id for multi-instance, by default all nodes will be used")
    group.add_argument("--use_logical_core", action='store_true', default=False,
                       help="Whether only use physical cores")
    group.add_argument("--disable_numactl", action='store_true', default=False,
                       help="Disable numactl")
    group.add_argument("--disable_taskset", action='store_true', default=False,
                       help="Disable taskset")
    group.add_argument("--core_list", metavar='\b', default=None, type=str,
                       help="Specify the core list as 'core_id, core_id, ....', otherwise, all the cores will be used.")
    group.add_argument("--log_path", metavar='\b', default="", type=str,
                       help="The log file directory. Default path is '', which means disable logging to files.")
    group.add_argument("--log_file_prefix", metavar='\b', default="run", type=str,
                       help="log file prefix")

def add_kmp_iomp_params(parser):

    group = parser.add_argument_group("IOMP Parameters")
    group.add_argument("--disable_iomp", action='store_true', default=False,
                       help="By default, we use Intel OpenMP and libiomp5.so will be add to LD_PRELOAD")


def parse_args():
    """
    Helper function parsing the command line options
    @retval ArgumentParser
    """
    parser = ArgumentParser(description="This is a script for launching PyTorch training and inference on Intel Xeon CPU "
                                        "with optimal configurations. Now, single instance inference/training, multi-instance "
                                        "inference/training and distributed training with oneCCL backend is enabled. "
                                        "To get the peak performance on Intel Xeon CPU, the script optimizes the configuration "
                                        "of thread and memory management. For thread management, the script configures thread "
                                        "affinity and the preload of Intel OMP library. For memory management, it configures "
                                        "NUMA binding and preload optimized memory allocation library (e.g. tcmalloc, jemalloc) "
                                        "\n################################# Basic usage ############################# \n"
                                        "\n 1. single instance\n"
                                        "\n   >>> python -m intel_extension_for_pytorch.cpu.launch python_script args \n"
                                        "\n2. multi-instance \n"
                                        "\n    >>> python -m intel_extension_for_pytorch.cpu.launch --ninstances xxx --ncore_per_instance xx python_script args\n"
                                        "\n3. Single-Node multi-process distributed training\n"
                                        "\n    >>> python  -m intel_extension_for_pytorch.cpu.launch --distributed  python_script args\n"
                                        "\n4. Multi-Node multi-process distributed training: (e.g. two nodes)\n"
                                        "\n   rank 0: *(IP: 192.168.10.10, and has a free port: 295000)*\n"
                                        "\n   >>> python -m intel_extension_for_pytorch.cpu.launch --distributed --nproc_per_node=2\n"
                                        "\n       --nnodes=2 --hostfile hostfile python_script args\n"
                                        "\n############################################################################# \n",
                                        formatter_class=RawTextHelpFormatter)

    parser.add_argument("--multi_instance", action='store_true', default=False,
                        help="Enable multi-instance, by default one instance per node")

    parser.add_argument('--distributed', action='store_true', default=False,
                        help='Enable distributed training.')
    parser.add_argument("-m", "--module", default=False, action="store_true",
                        help="Changes each process to interpret the launch script "
                             "as a python module, executing with the same behavior as"
                             "'python -m'.")

    parser.add_argument("--no_python", default=False, action="store_true",
                        help="Do not prepend the --program script with \"python\" - just exec "
                             "it directly. Useful when the script is not a Python script.")

    add_memory_allocator_params(parser)
    add_kmp_iomp_params(parser)

    add_distributed_training_params(parser)
    add_multi_instance_params(parser)
    # positional
    parser.add_argument("program", type=str,
                        help="The full path to the proram/script to be launched. "
                             "followed by all the arguments for the script")

    # rest from the training program
    parser.add_argument('program_args', nargs=REMAINDER)
    return parser.parse_args()

def main():

    env_before = set(os.environ.keys())
    if platform.system() == "Windows":
        raise RuntimeError("Windows platform is not supported!!!")

    args = parse_args()
    if args.log_path:
        path = os.path.dirname(args.log_path if args.log_path.endswith('/') else args.log_path + '/')
        if not os.path.exists(path):
            os.makedirs(path)
        args.log_path = path

        args.log_file_prefix = '{}_{}'.format(args.log_file_prefix, datetime.now().strftime("%Y%m%d%H%M%S"))
        fileHandler = logging.FileHandler("{0}/{1}_instances.log".format(args.log_path, args.log_file_prefix))
        logFormatter = logging.Formatter(format_str)
        fileHandler.setFormatter(logFormatter)
        logger.addHandler(fileHandler)

    if args.distributed and args.multi_instance:
        raise RuntimeError("Either args.distributed or args.multi_instance should be set")

    if args.latency_mode and args.throughput_mode:
        raise RuntimeError("Either args.latency_mode or args.throughput_mode should be set")

    if args.nnodes > 1:
        args.distributed = True

    if not args.no_python and not args.program.endswith(".py"):
        logger.error("For non Python script, you should use '--no_python' parameter.")
        exit()

    # Verify LD_PRELOAD
    if "LD_PRELOAD" in os.environ:
        lst_valid = []
        tmp_ldpreload = os.environ["LD_PRELOAD"]
        for item in tmp_ldpreload.split(":"):
            if item != "":
                matches = glob.glob(item)
                if len(matches) > 0:
                    lst_valid.append(item)
                else:
                    logger.warning("{} doesn't exist. Removing it from LD_PRELOAD.".format(item))
        if len(lst_valid) > 0:
            os.environ["LD_PRELOAD"] = ":".join(lst_valid)
        else:
            os.environ["LD_PRELOAD"] = ""

    launcher = None
    if args.distributed:
        launcher = DistributedTrainingLauncher()
    else:
        launcher = MultiInstanceLauncher()

    launcher.launch(args)
    for x in sorted(set(os.environ.keys()) - env_before):
        logger.debug('{0}={1}'.format(x, os.environ[x]))

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import warnings

from intel_extension_for_pytorch import optim

class _IPEXConvNd(nn.Module):
    __constants__ = ['stride', 'padding', 'dilation', 'groups',
                     'out_channels', 'kernel_size']

    def __init__(self, dense_module):
        super(_IPEXConvNd, self).__init__()
        self.out_channels = dense_module.out_channels
        self.in_channels = dense_module.in_channels
        self.kernel_size = dense_module.kernel_size
        self.stride = dense_module.stride
        self.padding = dense_module.padding
        self.dilation = dense_module.dilation
        self.groups = dense_module.groups
        self.prepack_input_shape = dense_module.input_shape if hasattr(dense_module, "input_shape") else []
        self.weight_channels_last = dense_module.weight.is_contiguous(memory_format=torch.channels_last) \
            or dense_module.weight.is_contiguous(memory_format=torch.channels_last_3d)

        # TODO: ".clone()" will make weight shared by multiple module not shared anymore
        # related issues: https://github.com/intel-innersource/frameworks.ai.pytorch.ipex-cpu/issues/65
        if dense_module.bias is not None:
            self.bias = nn.Parameter(
                dense_module.bias.detach().clone(),
                requires_grad = dense_module.bias.requires_grad)
            if hasattr(dense_module, 'master_bias'):
                self.master_bias = dense_module.master_bias
            elif hasattr(dense_module, 'bias_trail'):
                self.bias_trail = dense_module.bias_trail
        else:
            self.register_parameter('bias', None)
        # create conv op context
        self.ctx = torch.ops.ipex_prepack.convolution_prepack(
            dense_module.weight, self.bias, self.stride, self.padding,
            self.dilation, self.groups,
            self.weight_channels_last, self.prepack_input_shape
        )

        self.weight = nn.Parameter(self.ctx.get_weight(), requires_grad = dense_module.weight.requires_grad)

        # pack master_weight or weight_trail if needed
        if hasattr(dense_module, 'master_weight'):
            self.master_weight = self.ctx.pack(
                dense_module.master_weight.detach().clone()
            )
        elif hasattr(dense_module, 'weight_trail'):
            self.weight_trail = self.ctx.pack(
                dense_module.weight_trail.detach().clone(),
            )

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        assert not keep_vars, "can not using keep_vars true when to save _IPEXConvNd's parameters"
        if self.bias is not None:
            if hasattr(self, 'master_bias'):
                bias = self.master_bias
            elif hasattr(self, 'bias_trail'):
                bias = torch.ops.torch_ipex.cat_bfloat16_float(self.bias, self.bias_trail)
            else:
                bias = self.bias
            destination[prefix + 'bias'] = bias.detach()
        if hasattr(self, 'master_weight'):
            weight = self.master_weight
        elif hasattr(self, 'weight_trail'):
            weight = torch.ops.torch_ipex.cat_bfloat16_float(self.weight, self.weight_trail)
        else:
            weight = self.weight
        destination[prefix + 'weight'] = self.ctx.to_public(weight.detach())

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        assert False, "_IPEXConvNd does not support _load_from_state_dict method"

    def forward(self, x):
        return torch.ops.torch_ipex.convolution_forward(x, self.weight, self.bias, self.ctx.get_data_handle())

class _IPEXConv1d(_IPEXConvNd):
    def __init__(self, dense_module):
        super(_IPEXConv1d, self).__init__(dense_module)

class _IPEXConv2d(_IPEXConvNd):
    def __init__(self, dense_module):
        super(_IPEXConv2d, self).__init__(dense_module)

class _IPEXConv3d(_IPEXConvNd):
    def __init__(self, dense_module):
        super(_IPEXConv3d, self).__init__(dense_module)

class _IPEXLinear(torch.nn.Module):
    def __init__(self, dense_module):
        super(_IPEXLinear, self).__init__()

        # prepare batch size
        self.batch_size_collapsed = None
        if hasattr(dense_module, "input_shape"):
            self.batch_size_collapsed = 1
            for i in range(len(dense_module.input_shape) - 1):
                self.batch_size_collapsed *= dense_module.input_shape[i]

        # TODO:".clone()" will make weight shared by multiple module not shared anymore
        # related issues: https://github.com/intel-innersource/frameworks.ai.pytorch.ipex-cpu/issues/65
        if dense_module.bias is not None:
            self.bias = nn.Parameter(
                dense_module.bias.detach().clone(),
                requires_grad = dense_module.bias.requires_grad)
            if hasattr(dense_module, 'master_bias'):
                self.master_bias = dense_module.master_bias
            elif hasattr(dense_module, 'bias_trail'):
                self.bias_trail = dense_module.bias_trail
        else:
            self.register_parameter('bias', None)

        # create linear op context
        self.ctx = torch.ops.ipex_prepack.linear_prepack(
            dense_module.weight, self.bias, self.batch_size_collapsed
        )

        self.weight = nn.Parameter(self.ctx.get_weight(), requires_grad = dense_module.weight.requires_grad)

        # pack master_weight or weight_trail if needed
        if hasattr(dense_module, 'master_weight'):
            self.master_weight = self.ctx.pack(
                dense_module.master_weight.detach().clone()
            )
        elif hasattr(dense_module, 'weight_trail'):
            self.weight_trail = self.ctx.pack(
                dense_module.weight_trail.detach().clone(),
            )

    def forward(self, x):
        return torch.ops.torch_ipex.ipex_linear(
            x, self.weight, self.bias, self.ctx.get_data_handle()
        )

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        assert not keep_vars, "can not using keep_vars true when to save _IPEXLinear's parameters"
        if self.bias is not None:
            if hasattr(self, 'master_bias'):
                bias = self.master_bias
            elif hasattr(self, 'bias_trail'):
                bias = torch.ops.torch_ipex.cat_bfloat16_float(self.bias, self.bias_trail)
            else:
                bias = self.bias
            destination[prefix + 'bias'] = bias.detach()

        if hasattr(self, 'master_weight'):
            weight = self.master_weight
        elif hasattr(self, 'weight_trail'):
            weight = torch.ops.torch_ipex.cat_bfloat16_float(self.weight, self.weight_trail)
        else:
            weight = self.weight
        destination[prefix + 'weight'] = self.ctx.to_public(weight.detach())

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        assert False, "_IPEXLinear does not support _load_from_state_dict method"

class _IPEXConvTransposeNd(nn.Module):
    __constants__ = ['stride', 'padding', 'dilation', 'groups',
                     'out_channels', 'kernel_size', 'output_padding']

    def __init__(self, dense_module):
        super(_IPEXConvTransposeNd, self).__init__()
        self.out_channels = dense_module.out_channels
        self.in_channels = dense_module.in_channels
        self.kernel_size = dense_module.kernel_size
        self.stride = dense_module.stride
        self.padding = dense_module.padding
        self.groups = dense_module.groups
        self.dilation = dense_module.dilation
        self.output_padding = dense_module.output_padding
        self.prepack_input_shape = dense_module.input_shape if hasattr(dense_module, "input_shape") else []
        self.weight_channels_last = dense_module.weight.is_contiguous(memory_format=torch.channels_last) \
            or dense_module.weight.is_contiguous(memory_format=torch.channels_last_3d)

        if dense_module.bias is not None:
            self.bias = nn.Parameter(
                dense_module.bias.detach().clone(),
                requires_grad = dense_module.bias.requires_grad)
            if hasattr(dense_module, 'master_bias'):
                self.master_bias = dense_module.master_bias
            elif hasattr(dense_module, 'bias_trail'):
                self.bias_trail = dense_module.bias_trail
        else:
            self.register_parameter('bias', None)
        # create conv op context
        self.ctx = torch.ops.ipex_prepack.conv_transpose_prepack(
            dense_module.weight, self.bias, self.stride, self.padding,
            self.output_padding, self.groups, self.dilation,
            self.weight_channels_last, self.prepack_input_shape
        )

        self.weight = nn.Parameter(self.ctx.get_weight(), requires_grad = dense_module.weight.requires_grad)

        # pack master_weight or weight_trail if needed
        if hasattr(dense_module, 'master_weight'):
            self.master_weight = self.ctx.pack(
                dense_module.master_weight.detach().clone())
        elif hasattr(dense_module, 'weight_trail'):
            self.weight_trail = self.ctx.pack(
                dense_module.weight_trail.detach().clone()
        )

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        assert not keep_vars, "can not using keep_vars true when to save _IPEXConvTransposeNd's parameters"
        if self.bias is not None:
            if hasattr(self, 'master_bias'):
                bias = self.master_bias
            elif hasattr(self, 'bias_trail'):
                bias = torch.ops.torch_ipex.cat_bfloat16_float(self.bias, self.bias_trail)
            else:
                bias = self.bias
            destination[prefix + 'bias'] = bias.detach()
        if hasattr(self, 'master_weight'):
            weight = self.master_weight
        elif hasattr(self, 'weight_trail'):
            weight = torch.ops.torch_ipex.cat_bfloat16_float(self.weight, self.weight_trail)
        else:
            weight = self.weight
        destination[prefix + 'weight'] = self.ctx.to_public(weight.detach())

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        assert False, "_IPEXConvTransposeNd does not support _load_from_state_dict method"

    def forward(self, x):
        return torch.ops.torch_ipex.conv_transpose(
            x,
            self.weight,
            self.bias,
            self.ctx.get_data_handle())

class _IPEXConvTranspose2d(_IPEXConvTransposeNd):
    def __init__(self, dense_module):
        super(_IPEXConvTranspose2d, self).__init__(dense_module)

class _IPEXConvTranspose3d(_IPEXConvTransposeNd):
    def __init__(self, dense_module):
        super(_IPEXConvTranspose3d, self).__init__(dense_module)

IPEX_WEIGHT_PREPACK_MODULE = {
    torch.nn.Linear: _IPEXLinear,
    torch.nn.Conv2d: _IPEXConv2d,
    torch.nn.Conv3d: _IPEXConv3d,
    torch.nn.Conv1d: _IPEXConv1d,
    torch.nn.ConvTranspose2d: _IPEXConvTranspose2d,
    torch.nn.ConvTranspose3d: _IPEXConvTranspose3d,
}

def _should_prepack(module, auto_kernel_selection):
    if type(module) not in IPEX_WEIGHT_PREPACK_MODULE:
        return False
    # If hook is on `weight` or `bias`, will not prepack.
    if module._forward_pre_hooks is not None:
        for _, hook in module._forward_pre_hooks.items():
            if hook.name == 'weight' or hook.name == 'bias':
                return False
    if module._forward_hooks is not None:
        for _, hook in module._forward_hooks.items():
            if hook.name == 'weight' or hook.name == 'bias':
                return False
    if module._backward_hooks is not None:
        for _, hook in module._backward_hooks.items():
            if hook.name == 'weight' or hook.name == 'bias':
                return False
    if isinstance(module, torch.nn.Linear) and not auto_kernel_selection and module.weight.dtype is torch.float:
        # For now we simply distinguish "mkl" and "mkldnn" backend by "weight prepack"
        # Does not prepack Linear for FP32 to choose "mkl" backend
        return False
    if isinstance(module, torch.nn.ConvTranspose2d):
        if module.padding[0] - module.output_padding[0] + module.stride[0] <= 0:
            return False
        if module.padding[1] - module.output_padding[1] + module.stride[1] <= 0:
            return False
    if isinstance(module, torch.nn.ConvTranspose3d):
        if module.padding[0] - module.output_padding[0] + module.stride[0] <= 0:
            return False
        if module.padding[1] - module.output_padding[1] + module.stride[1] <= 0:
            return False
        if module.padding[2] - module.output_padding[2] + module.stride[2] <= 0:
            return False
    return True

def weight_prepack_with_ipex(module, optimizer, params_attr, auto_kernel_selection):
    def convert(m, optimizer, params_attr, auto_kernel_selection):
        if _should_prepack(m, auto_kernel_selection) and (m.weight.dtype == torch.float32 or m.weight.dtype == torch.bfloat16):
            weight = m.master_weight if hasattr(m, "master_weight") else m.weight
            if weight not in params_attr:
                params_attr[weight] = {}
            new_m = IPEX_WEIGHT_PREPACK_MODULE[type(m)](m)
            params_attr[weight].update({
                'op': type(m),
                'ctx': new_m.ctx})
            if hasattr(new_m, "weight_channels_last"):
                params_attr[weight]['weight_channels_last'] = new_m.weight_channels_last,
            if 'bf16_param' in params_attr[weight]:
                params_attr[weight]['bf16_param'] = new_m.weight
            elif 'trail' in params_attr[weight]:
                params_attr[weight]['trail'] = new_m.weight_trail
            # update entry from origin weight to packed weight, from origin bias to cloned bias
            new_weight = new_m.master_weight if hasattr(m, "master_weight") else new_m.weight
            params_attr[new_weight] = params_attr.pop(weight)
            params_pair = {weight: new_weight}
            if hasattr(m, 'bias') and m.bias is not None:
                bias = m.master_bias if hasattr(m, "master_bias") else m.bias
                new_bias = new_m.master_bias if hasattr(m, "master_bias") else new_m.bias
                params_pair.update({bias: new_bias})
                if bias in params_attr:
                    if 'bf16_param' in params_attr[bias]:
                        params_attr[bias]['bf16_param'] = new_m.bias
                    elif 'trail' in params_attr[bias]:
                        params_attr[bias]['trail'] = new_m.bias_trail
                    if bias in params_attr:
                        params_attr[new_bias] = params_attr.pop(bias)
            # replace optimizer's param with prepacked param, also prepack its state.
            optim._optimizer_utils.pack_optimizer_params_and_states(
                optimizer, params_pair, params_attr, m.weight.dtype)
            return new_m
        else:
            return m

    def convert_rec(m, optimizer, params_attr, auto_kernel_selection):
        new_m = convert(m, optimizer, params_attr, auto_kernel_selection)
        for name, sub_m in m.named_children():
            setattr(new_m, name, convert_rec(sub_m, optimizer, params_attr, auto_kernel_selection)[0])
        return new_m, optimizer, params_attr

    opt_model, opt_optmizer, params_attr = convert_rec(module, optimizer, params_attr, auto_kernel_selection)
    if opt_optmizer is not None:
        setattr(opt_optmizer, 'params_attr', params_attr)
        optim._optimizer_utils.patch_load_state_dict(opt_optmizer)
        optim._optimizer_utils.patch_state_dict(opt_optmizer)
    return opt_model, opt_optmizer, params_attr

def record_input_shape_for_prepack(module, sample_input):

    def hook_function(self, input):
        # input for linear/conv/transpose conv received here will be Tuple[Tensor]
        self.input_shape = input[0].shape

    def register_hook_function(module):
        if type(module) in [torch.nn.Linear, torch.nn.Conv2d, torch.nn.ConvTranspose2d]:
            module.register_forward_pre_hook(hook_function)

    def register_hook_function_rec(module):
        register_hook_function(module)
        for child in module.children():
            register_hook_function_rec(child)

    hook_function.name = "input_shape"
    register_hook_function_rec(module)
    module(*sample_input)



import torch
import intel_extension_for_pytorch._C as core

# Expand torch cpu autocast to support more data types in the future
class _autocast(torch.cpu.amp.autocast):
    def __enter__(self):
        self.prev_cache_enabled = torch.is_autocast_cache_enabled()
        self.prev = torch.is_autocast_cpu_enabled()
        self.prev_fast_dtype = core.get_autocast_dtype()
        torch.set_autocast_cpu_enabled(self._enabled)
        core.set_autocast_dtype(self.fast_dtype)
        torch.autocast_increment_nesting()
        torch.set_autocast_cache_enabled(self._cache_enabled)

    def __exit__(self, *args):
        # Drop the cache when we exit to a nesting level that's outside any instance of autocast.
        if torch.autocast_decrement_nesting() == 0:
            core.clear_autocast_cache()
            torch.clear_autocast_cache()
        torch.set_autocast_cpu_enabled(self.prev)
        core.set_autocast_dtype(self.prev_fast_dtype)
        torch.set_autocast_cache_enabled(self.prev_cache_enabled)
        return False


torch.cpu.amp.autocast = _autocast



import torch
import torch.nn as nn
from intel_extension_for_pytorch.optim import _optimizer_utils, _lamb
import types

# IPEX does not cast all module parameters for acc reason, such as BN
IPEX_WEIGHT_CAST_MODULE = {
    # align with auto cast white list
    torch.nn.Linear,
    torch.nn.Conv1d,
    torch.nn.Conv2d,
    torch.nn.Conv3d,
    torch.nn.ConvTranspose1d,
    torch.nn.ConvTranspose2d,
    torch.nn.ConvTranspose3d,
    # ipex support
    torch.nn.EmbeddingBag,
    torch.nn.Embedding,
}

def _save_to_state_dict(self, destination, prefix, keep_vars):
    # cast weight
    temp_weight = self.weight
    if self.master_weight_split:
        self.weight = torch.nn.Parameter(
            torch.ops.torch_ipex.cat_bfloat16_float(self.weight.data, self.weight_trail),
            requires_grad=temp_weight.requires_grad)
    else:
        self.weight = torch.nn.Parameter(
            self.master_weight,
            requires_grad=temp_weight.requires_grad)
    # cast bias
    if hasattr(self, 'bias') and self.bias is not None:
        temp_bias = self.bias
        if self.master_weight_split:
            self.bias = torch.nn.Parameter(
                torch.ops.torch_ipex.cat_bfloat16_float(self.bias.data, self.bias_trail),
                requires_grad=temp_bias.requires_grad)
        else:
            self.bias = torch.nn.Parameter(
                self.master_bias,
                requires_grad=temp_bias.requires_grad)
    super(type(self), self)._save_to_state_dict(destination, prefix, keep_vars)
    self.weight = temp_weight
    if hasattr(self, 'bias') and self.bias is not None:
        self.bias = temp_bias

def weight_dtype_convert_with_ipex(module, optimizer, params_attr, master_weight_split):

    def cast_attr(m, attr, master_weight_split, params_attr, optimizer):
        # cast weight/bias for BF16 dtype
        float_param = getattr(m, attr)
        params_attr[float_param] = {}
        if master_weight_split:
            top_half, bot_half = torch.ops.torch_ipex.split_float_bfloat16(float_param.data)
            setattr(m, attr + '_trail', bot_half)
            setattr(m, attr, nn.Parameter(top_half.detach(), requires_grad=float_param.requires_grad))
            params_attr[float_param]['trail'] = getattr(m, attr + '_trail')
        else:
            setattr(m, 'master_' + attr, float_param.data)
            setattr(m, attr, nn.Parameter(float_param.detach().bfloat16(), requires_grad=float_param.requires_grad))
            params_attr[float_param]['bf16_param'] = getattr(m, attr)
        # update attr entry, always use params in optimzer as "key"
        # while master weight split, key is m.weight/bias, if not split, key is m.master_weight/master_bias
        attr_name = attr if master_weight_split else 'master_' + attr
        params_attr[getattr(m, attr_name)] = params_attr.pop(float_param)
        _optimizer_utils.refresh_optimizer_params_after_cast(m, attr, float_param, master_weight_split, optimizer)

    def convert(m):
        if type(m) in IPEX_WEIGHT_CAST_MODULE:
            setattr(m, 'master_weight_split', master_weight_split)
            # replace weight
            cast_attr(m, 'weight', master_weight_split, params_attr, optimizer)
            if hasattr(m, 'bias') and m.bias is not None:
                # replace bias
                cast_attr(m, 'bias', master_weight_split, params_attr, optimizer)
            # for resume training reason, we always save float tensors
            # replace module method to ensure return float params while call "state_dict()"
            setattr(m, '_save_to_state_dict', types.MethodType(_save_to_state_dict, m))
        return m

    def convert_rec(m):
        new_m = convert(m)
        for name, sub_m in m.named_children():
            setattr(new_m, name, convert_rec(sub_m))
        return new_m

    casted_model, casted_optimizer, params_attr = convert_rec(module), optimizer, params_attr

    if optimizer is not None:
        _optimizer_utils.patch_load_state_dict(casted_optimizer)
        setattr(casted_optimizer, 'params_attr', params_attr)
        if not master_weight_split:
            _optimizer_utils.patch_step_for_master_weight_training(casted_optimizer)
            _optimizer_utils.patch_zero_grad_for_master_weight_training(casted_optimizer)

    return casted_model, casted_optimizer, params_attr



import torch
from torch import nn

class FrozenBatchNorm2d(nn.Module):
    r"""
    BatchNorm2d where the batch statistics and the affine parameters are fixed

    Args:
        num_features (int): :math:`C` from an expected input of size :math:`(N, C, H, W)`

    Shape
        - Input: :math:`(N, C, H, W)`
        - Output: :math:`(N, C, H, W)` (same shape as input)
    """

    def __init__(self,
        num_features: int,
        eps: float = 1e-5,
    ):
        super(FrozenBatchNorm2d, self).__init__()
        self.eps = eps
        self.register_buffer("weight", torch.ones(num_features))
        self.register_buffer("bias", torch.zeros(num_features))
        self.register_buffer("running_mean", torch.zeros(num_features))
        self.register_buffer("running_var", torch.ones(num_features))

    def forward(self, input):
        return torch.ops.torch_ipex.frozen_batch_norm(input, self.weight, self.bias, self.running_mean, self.running_var, self.eps)



import torch
import intel_extension_for_pytorch._C as core
import warnings

torch_embedding_bag = torch.embedding_bag

def _embeddingbag(weights, indices, offsets, scale_grad_by_freq, mode, sparse, per_sample_weights, include_last_offset, padding_idx):
    if core.embedding_bag_fast_path_sum(weights, per_sample_weights, mode, padding_idx):
        ret = torch.ops.torch_ipex.embedding_bag(weights, indices, offsets, sparse, include_last_offset)
        # torch.embedding_bag expected 4 Tensor returned
        # here we only return 1 tensor since the other three tensors are not needed in our fast path
        ret = [ret, torch.Tensor(), torch.Tensor(), torch.Tensor()]
    else:
        warnings.warn('Fallback to torch.embedding bag')
        ret = torch_embedding_bag(weights, indices, offsets, scale_grad_by_freq, mode, sparse, per_sample_weights, include_last_offset, padding_idx)
    return ret

torch.embedding_bag = _embeddingbag



import json
import os
import torch
import torch.nn as nn
import torch.nn.functional as F
from intel_extension_for_pytorch.nn.functional import interaction

from ._utils import ParentNode, set_node_output_quantized

add_inplace_ops = [str(torch.Tensor.add_)]
add_ops = [str(torch.add), str(torch.Tensor.add)]
elt_wise_q_ops = [str(torch.Tensor.relu), str(torch.relu), str(F.relu), str(nn.ReLU)]
elt_wise_noq_ops = [str(torch.relu_), str(torch.sigmoid_), str(nn.ReLU), str(torch.Tensor.relu_), str(torch.Tensor.sigmoid_), \
    str(torch.nn.Hardtanh), str(F.hardtanh), str(F.hardtanh_),str(torch.nn.ELU), str(F.elu), str(F.elu_), \
        str(nn.SiLU), str(F.silu), str(torch.Tensor.sigmoid), str(torch.sigmoid), str(F.sigmoid), str(nn.Sigmoid), str(F.gelu), str(nn.GELU)]
conv_gemm_ops = [str(F.conv2d), str(nn.Conv2d), str(F.conv3d), str(nn.Conv3d), str(torch.conv2d), str(torch.conv3d), \
    str(F.conv_transpose2d), str(torch.nn.ConvTranspose2d), str(F.conv_transpose3d), str(torch.nn.ConvTranspose3d),
    str(torch.conv_transpose2d), str(torch.conv_transpose2d), str(F.linear), str(nn.Linear), str(torch.matmul), str(torch.Tensor.matmul)]
rnn_ops = [str(torch.nn.LSTM)]

# Those ops only support s8->s8 path, and also require the qscheme is per_tensor_symmetric.
s8_s8_symmetric_ops = [str(interaction), str(torch.ops.torch_ipex.interaction_forward), str(torch.embedding_bag), \
    str(F.embedding_bag), str(torch.nn.EmbeddingBag)]
conv_gemm_fs = [str(F.conv2d), str(F.conv3d), str(F.conv_transpose2d), str(F.conv_transpose3d), str(torch.conv2d), str(torch.conv3d), \
    str(torch.conv_transpose2d), str(torch.conv_transpose2d), str(F.linear), str(torch._C._nn.linear)]

def _default_recipe_init(nodes):
    r"""
    This function is about init default recipe: setting the quantizable op's inf dtype to qint8 or quint8 according the qconfig,
    there have some special cases, for some ops(interaction, EmbeddingBag), we only support some special quantization path, so if the related qconfig
    doesn't meet the requirements, we will not set their inf dtype.
    """
    for node in nodes:
        if isinstance(node, ParentNode):
            continue
        if node.qconfig is not None:
            # Add q+dq before the quantizable op firstly.
            for idx, tensor_info in enumerate(node.input_tensor_infos):
                # only support fp32 tensor->int8 tensor
                if tensor_info is not None and (tensor_info.orig_dtype == torch.float32) and tensor_info.id in node.input_scale_zero:
                    # gemm's weight
                    if node.type in conv_gemm_fs and idx == 1:
                        tensor_info.inf_dtype = node.qconfig.weight().dtype
                    else:
                        tensor_info.inf_dtype = node.qconfig.activation().dtype
                    node.input_tensor_force_inf_dtype[idx] = tensor_info.inf_dtype
            # For EmbeddingBag and interaction, we need to check the qconfig's setting, if not meet the requirements, reset the inputs'(or weight) inf dtype
            for tensor_info in node.weight_tensor_infos:
                # nn.EmbeddingBag use activation observer and only support torch.qint8 and torch.per_tensor_symmetric
                if tensor_info is not None and (tensor_info.orig_dtype == torch.float32) and (str(node.idx) + "_" + str(tensor_info.id) in node.weight_scale_zero):
                    if (node.type == str(torch.nn.EmbeddingBag) and node.qconfig.activation().dtype == torch.qint8 and \
                        node.qconfig.activation().qscheme == torch.per_tensor_symmetric) or node.type != str(torch.nn.EmbeddingBag):
                        tensor_info.inf_dtype = node.qconfig.weight().dtype
            # interaction only supports qint8 and torch.per_tensor_symmetric, if not meet the requirement,
            # reset the input's inf dtype.
            if node.type in s8_s8_symmetric_ops:
                if not(node.qconfig.activation().dtype == torch.qint8 and node.qconfig.activation().qscheme == torch.per_tensor_symmetric):
                    for idx, tensor_info in enumerate(node.input_tensor_infos):
                        if tensor_info is not None:
                            tensor_info.inf_dtype = tensor_info.orig_dtype
                            node.input_tensor_force_inf_dtype[idx] = tensor_info.inf_dtype

#TODO: making fusion pattern check more general.
def _find_fused_node_with_cur_elt_wise(node, ops):
    r"""
    Find a node before cur elt_wise which can be fused with cur elt_wise, which used by check
    whether has a op can be fused with elt_wise.
    """
    if len(node.pre_nodes) == 0:
        return None
    pre_node = node.pre_nodes[0]
    if pre_node is not None:
        if pre_node.type in ops:
            if len(pre_node.post_nodes) == 1:
                return pre_node
            elif len(node.post_nodes) == 1 and _find_conv_or_gemm_swish_fusion_node(node.post_nodes[0]):
                # conv+sigmoid+mul
                return pre_node
            else:
                return None
        elif pre_node.type in ([str(nn.Identity)] + elt_wise_q_ops + elt_wise_noq_ops) and \
            len(pre_node.post_nodes) == 1 and len(pre_node.pre_nodes) > 0:
            return _find_fused_node_with_cur_elt_wise(pre_node.pre_nodes[0], ops)
        else:
            return None
    else:
        return None

def _find_fused_node_with_cur_add(node, ops):
    r"""
    Find a node before the cur node which can be fused with cur add node, which used to check
    whether has a node can be fused with add.
    """
    if len(node.pre_nodes) == 0:
        return None
    if len(node.pre_nodes) > 0:
        if node.pre_nodes[0].type in ops and len(node.pre_nodes[0].post_nodes) == 1 and node.pre_nodes[0].qconfig is not None:
            return node.pre_nodes[0]
        elif node.pre_nodes[0].type == str(nn.Identity) and \
            len(node.pre_nodes[0].post_nodes) == 1 and len(node.pre_nodes[0].pre_nodes) > 0:
            fused_node =  _find_fused_node_with_cur_add(node.pre_nodes[0], ops)
            if fused_node is not None:
                return node.pre_nodes[0]
            else:
                return None

        if len(node.pre_nodes) == 2:
            if node.pre_nodes[1].type in ops and len(node.pre_nodes[1].post_nodes) == 1 and node.pre_nodes[1].qconfig is not None:
                return node.pre_nodes[1]
            elif node.pre_nodes[1].type == str(nn.Identity) and \
                len(node.pre_nodes[1].post_nodes) == 1 and len(node.pre_nodes[1].pre_nodes) > 0:
                fused_node =  _find_fused_node_with_cur_add(node.pre_nodes[1], ops)
                if fused_node is not None:
                    return node.pre_nodes[1]
                else:
                    return None
        return None

def _find_conv_or_gemm_swish_fusion_node(node):
    r"""
    Check whether has conv/gemm_sigmoid_mul fusion before cur node(including).
        conv/gemm
          /  \
         /  sigmoid
         \     /
           mul(_)
    """
    mul_ops =  [str(torch.mul), str(torch.Tensor.mul), str(torch.Tensor.mul_)]
    sigmoid_ops = [str(torch.Tensor.sigmoid), str(torch.Tensor.sigmoid_), str(torch.sigmoid), str(torch.sigmoid_),\
        str(F.sigmoid), str(torch.nn.Sigmoid)]
    if node.type in mul_ops and len(node.pre_nodes) == 2:
        if node.pre_nodes[0].type in conv_gemm_ops and node.pre_nodes[1].type in sigmoid_ops:
            if len(node.pre_nodes[0].post_nodes) == 2 and len(node.pre_nodes[1].post_nodes) == 1 and \
                node.pre_nodes[1] in node.pre_nodes[0].post_nodes:
                return node.pre_nodes[0]
        elif node.pre_nodes[1].type in conv_gemm_ops and node.pre_nodes[0].type in sigmoid_ops:
            if len(node.pre_node[1].post_nodes) == 2 and len(node.pre_nodes[0].post_nodes) == 1 and \
                node.pre_nodes[0] in node.pre_node[1].post_nodes:
                return node.pre_nodes[1]
    return None

def _check_has_quantizable_node_before_node(node):
    r"""
    This function is about check whether has a quantizable node before(including) the given node,
    which is used to check whether insert fake quant before one quantizable node or not. For example,
    given_node->quantizable_node, if the given node is a none-quantizable node(also not a fusion groups nodes),
    we can avoid inserting fake quant before this quantizable node.
    """
    if node.type == str(nn.Identity):
        if len(node.pre_nodes) > 0:
            return _check_has_quantizable_node_before_node(node.pre_nodes[0])
        else:
            return False
    else:
        # check whether has a qconfig
        if node.qconfig is None:
            if len(node.pre_nodes) == 0:
                return False
            # conv/gemm+add(_)+elt_wise
            if node.type in elt_wise_noq_ops:
                fused_elt_wise_node = _find_fused_node_with_cur_elt_wise(node, conv_gemm_ops + add_ops + add_inplace_ops)
                if fused_elt_wise_node is not None:
                    # if fused_elt_wise_node is add_inplace_op, make sure it can also fused with conv/gemm.
                    if fused_elt_wise_node.type in add_inplace_ops:
                        fused_add_node = _find_fused_node_with_cur_add(node, conv_gemm_ops)
                        if  fused_add_node is not None and fused_add_node.qconfig is not None:
                            return True
                        else:
                            return False
                    else:
                        if fused_elt_wise_node.qconfig is not None:
                            return True
                        else:
                            return False
            elif node.type in add_inplace_ops: # check gemm+add_
                fused_add_wise_node = _find_fused_node_with_cur_add(node, conv_gemm_ops)
                if fused_add_wise_node is not None and fused_add_wise_node.qconfig is not None:
                    return True
            # conv+sigmoid+mul(_)
            fused_conv_or_gemm_swish_node = _find_conv_or_gemm_swish_fusion_node(node)
            if fused_conv_or_gemm_swish_node is not None and fused_conv_or_gemm_swish_node.qconfig is not None:
                return True
            return False
        else:
            if node.type in s8_s8_symmetric_ops:
                if node.type in [str(interaction), str(torch.ops.torch_ipex.interaction_forward)]:
                    for force_inf_dtype in node.input_tensor_force_inf_dtype:
                        if force_inf_dtype.inf_dtype == torch.qint8:
                            return True
                    return False
                else:
                    # EmbeddingBag
                    if node.weight_tensor_infos[0].inf_dtype == torch.qint8:
                        return True
                    else:
                        return False
            else:
                # for none ipex customer op, if have a qconfig, we can say it is a quantizable op.
                return True

def _add_recipe(node):
    '''
    Case1: add has pre gemm node.
    Given  gemm     op             gemm         op                gemm       op
             \     /                 \         /                   \       /
              \   /       ==>    fake_quant (fake_quant?)     ==>   \   (fake_quant?)
               \ /                     \    /                        \   /
               add                       add                          add

          gemm     fp32_op          gemm     quantizable_op
    ==>    \        /                \         /
            \      /          or      \     fake_quant
             \    /                    \    /
              add                       add

    Case2: add doesn't have pre conv/gemm node.
    For this case, if one add input has one none-quantizable op, we will don't insert fake quant before it.
    '''
    conv_gemm_node = _find_fused_node_with_cur_add(node, conv_gemm_ops)
    if conv_gemm_node is None:
        #  If pre_nodes don't have gemm node, need to check whether have quantizable node before it,
        #  if does't have quantizable node before it, we will not insert fake quant before add.
        # hoping all input nodes are quantizable node.
        if len(node.pre_nodes) > 0:
            add_1_has_pre_quantizable_op = _check_has_quantizable_node_before_node(node.pre_nodes[0])
            add_2_has_pre_quantizable_op = False
            if len(node.pre_nodes) == 2:
                add_2_has_pre_quantizable_op = _check_has_quantizable_node_before_node(node.pre_nodes[1])
            if not (add_1_has_pre_quantizable_op and add_2_has_pre_quantizable_op):
                for idx, tensor_info in enumerate(node.input_tensor_infos):
                    tensor_info.inf_dtype = tensor_info.orig_dtype
                    node.input_tensor_force_inf_dtype[idx] = tensor_info.inf_dtype
        else:
            for idx, tensor_info in enumerate(node.input_tensor_infos):
                tensor_info.inf_dtype = tensor_info.orig_dtype
                node.input_tensor_force_inf_dtype[idx] = tensor_info.inf_dtype
    else:
        # add can fused with gemm.
        if node.input_tensor_infos[0] is not None and node.input_tensor_infos[0] in conv_gemm_node.output_tensor_infos:
            node.input_tensor_infos[0].inf_dtype = node.input_tensor_infos[0].orig_dtype
            node.input_tensor_force_inf_dtype[0] = node.input_tensor_infos[0].inf_dtype
            '''
            # TODO: set another input's dtype, if another's input is from non-quantizable op, we can remove the fake quant.
            # now if for quantized conv/gemm_add, the input's should be all INT8.
            if node.input_tensor_infos[1] is not None:
                if node.input_tensor_infos[1] in node.pre_nodes[0].output_tensor_infos:
                    pre_node = node.pre_nodes[0]
                elif len(node.pre_nodes) == 2 and node.input_tensor_infos[1] in node.pre_nodes[1].output_tensor_infos:
                    pre_node = node.pre_nodes[1]
                else:
                    pre_node = None
                if pre_node is not None:
                    add_quantize_add_2 = _check_has_quantizable_node_before_node(pre_node)
                else:
                    add_quantize_add_2 = False
                if not add_quantize_add_2:
                    node.input_tensor_infos[1].inf_dtype = node.input_tensor_infos[1].orig_dtype
                    node.input_tensor_force_inf_dtype[1] = node.input_tensor_infos[1].inf_dtype
              '''
        elif node.input_tensor_infos[1] is not None and node.input_tensor_infos[1] in conv_gemm_node.output_tensor_infos:
            node.input_tensor_infos[1].inf_dtype = node.input_tensor_infos[1].orig_dtype
            node.input_tensor_force_inf_dtype[1] = node.input_tensor_infos[1].inf_dtype
            '''
            # TODO: set another input's dtype, if another's input is from non-quantizable op, we can remove the fake quant.
            # now if for quantized conv/gemm_add, the input's should be all INT8.
            if node.input_tensor_infos[0] is not None:
                if node.input_tensor_infos[0] in node.pre_nodes[0].output_tensor_infos:
                    pre_node = node.pre_nodes[0]
                elif len(node.pre_nodes) == 2 and node.input_tensor_infos[0] in node.pre_nodes[1].output_tensor_infos:
                    pre_node = node.pre_nodes[1]
                else:
                    pre_node = None
                if pre_node is not None:
                    add_quantize_add_1 = _check_has_quantizable_node_before_node(pre_node)
                else:
                    add_quantize_add_1 = False
                if not add_quantize_add_1:
                    node.input_tensor_infos[0].inf_dtype = node.input_tensor_infos[0].orig_dtype
                    node.input_tensor_force_inf_dtype[0] = node.input_tensor_infos[0].inf_dtype
            '''

# get a default recipe
def get_default_recipe(nodes):
    r"""
    This function is about get default recipe which set where fake quant is inserted for the quantizable ops.
    """
    # step1: Quantization state init. Quantize inputs before quantizable node by setting their input's inf_dtype to 
    # qconfig.activation().dtype, and also setting the weight's inf_dtype to 
    # qconfig.weight().dtype if a module has a weight.
    _default_recipe_init(nodes)
    # step2: Optimization
    # 1. For conv, gemm, and LSTM,  we always quantize its' inputs and weight, so we keep them state.
    #    and for embedding_bag, which only has a weight, we always quantize it's weight to
    #    save memory space and bandwidth, we also keep it's state.
    # 2. For remaining quantizable ops (pooling, elt-wise op and add) which meet the following requirements, we will
    # update them inputs' quantization state.
    #   1. If it is a part of a quantized fusion pattern, don't need to quantize any inputs from inside the pattern.
    #   2. If any of its inputs outside the fusion pattern are from non-quantized op, don't quantize all inputs outside the pattern.
    #   3. If it is not part of a quantized fusion pattern, don't quantize all inputs if its one input from non-quantized op.
    # 3. For quantizable ops (pooling, relu, flatten, interation and embedding) forcing quantized output, need to quantize its output if it is quantized.
    # 4. For interation and embedding, we only support s8->s8 symmetric quantization, so if doesn't meet the requiresments, don't need to quantize its inputs.
    # Note: the fusion pattern we are supported is conv/gemm/add + elt-wise, conv/gemm + add, conv/gemm + add + elt-wise.
    # which means some ops can be combined with a single op to compute, but they are mathematically equivalent.
    embedding_bag_ops = [str(torch.embedding_bag), str(F.embedding_bag), str(torch.nn.EmbeddingBag)]
    for node in nodes:
        if isinstance(node, ParentNode):
            continue
        if node.qconfig is not None and not node.type in (conv_gemm_ops + rnn_ops + embedding_bag_ops):
            if node.type in add_ops:
                # gemm+add fusion
                _add_recipe(node)
            elif node.type in  elt_wise_q_ops:
                # don't have a pre_node, we can say it doesn't have a pre quantizable node.
                has_pre_quantized_node = True
                # If Has gemm(add) pre_op can be fused, not insert fake quant.
                if len(node.pre_nodes) > 0:
                    if _find_fused_node_with_cur_elt_wise(node, conv_gemm_ops + add_ops + add_inplace_ops) is not None:
                        has_pre_quantized_node = False
                    else:
                        has_pre_quantized_node = _check_has_quantizable_node_before_node(node.pre_nodes[0])
                else:
                    has_pre_quantized_node = False
                if not has_pre_quantized_node:
                    node.input_tensor_infos[0].inf_dtype = node.input_tensor_infos[0].orig_dtype
                    node.input_tensor_force_inf_dtype[0] = node.input_tensor_infos[0].inf_dtype
            else:
                # For other quantizable node, we don't need add fake quant before it if it's pre node is one none-quantizable op.
                # Now all other quantizable node only have one input info, so we can check the one pre input node info to check
                # whether has a pre quantizable node.
                has_pre_quantized_node = True
                if len(node.pre_nodes) == 1:
                    has_pre_quantized_node = _check_has_quantizable_node_before_node(node.pre_nodes[0])
                elif len(node.pre_nodes) == 0:
                    has_pre_quantized_node = False
                # the node's pre node doesn't support int8 output.
                if not has_pre_quantized_node:
                    node.input_tensor_infos[0].inf_dtype = node.input_tensor_infos[0].orig_dtype
                    node.input_tensor_force_inf_dtype[0] = node.input_tensor_infos[0].inf_dtype

    set_node_output_quantized(nodes)



import torch
from functools import wraps

# Wrap torch.jit.trace to disable autocast cache when using torch.jit.trace
# within the scope of torch.cpu.amp.autocast.
# See https://github.com/pytorch/pytorch/pull/63552 for more information.
def disable_autocast_cache(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        prev = torch.is_autocast_cache_enabled()
        # Disable autocast cache
        if torch.is_autocast_cpu_enabled():
            torch.set_autocast_cache_enabled(False)
        traced = f(*args, **kwargs)
        torch.set_autocast_cache_enabled(prev)
        return traced
    return wrapper

torch.jit.trace = disable_autocast_cache(torch.jit.trace)



import torch
import functools
import warnings
import numpy as np
import intel_extension_for_pytorch as ipex
from .runtime_utils import get_core_list_of_node_id

class CPUPool(object):
    r"""
    An abstraction of a pool of CPU cores used for intra-op parallelism.

    Args:
        core_ids (list): A list of CPU cores' ids used for intra-op parallelism.
        node_id (int): A numa node id with all CPU cores on the numa node.
            ``node_id`` doesn't work if ``core_ids`` is set.

    Returns:
        intel_extension_for_pytorch.cpu.runtime.CPUPool: Generated
        intel_extension_for_pytorch.cpu.runtime.CPUPool object.
    """

    def __init__(self, core_ids: list = None, node_id: int = None):
        if core_ids is not None:
            if node_id is not None:
                warnings.warn("Both of core_ids and node_id are inputed. core_ids will be used with priority.")
            if type(core_ids) is range:
                core_ids = list(core_ids)
            assert type(core_ids) is list, "Input of core_ids must be the type of list[Int]"
            self.core_ids = core_ids
        else:
            assert node_id is not None, "Neither core_ids or node_id has been implemented"
            self.core_ids = get_core_list_of_node_id(node_id)
        self.cpu_pool = ipex._C.CPUPool(self.core_ids)

class pin(object):
    r"""
    Apply the given CPU pool to the master thread that runs the scoped code
    region or the function/method def.

    Args:
        cpu_pool (intel_extension_for_pytorch.cpu.runtime.CPUPool):
            intel_extension_for_pytorch.cpu.runtime.CPUPool object, contains
            all CPU cores used by the designated operations.

    Returns:
        intel_extension_for_pytorch.cpu.runtime.pin: Generated
        intel_extension_for_pytorch.cpu.runtime.pin object which can be used
        as a `with` context or a function decorator.
    """

    def __init__(self, cpu_pool: CPUPool):
        self.cpu_pool = cpu_pool
        ipex._C.init_runtime_ext()

    def __enter__(self):
        assert type(self.cpu_pool) is CPUPool
        self.previous_cpu_pool = ipex._C.get_current_cpu_pool()
        ipex._C.pin_cpu_cores(self.cpu_pool.core_ids)

    def __exit__(self, *args):
        ipex._C.set_cpu_pool(self.previous_cpu_pool)

    # Support decorator
    def __call__(self, func):
        @functools.wraps(func)
        def decorate_pin(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return decorate_pin

def is_runtime_ext_enabled():
    r"""
    Helper function to check whether runtime extension is enabled or not.

    Args:
       None (None): None

    Returns:
        bool: Whether the runtime exetension is enabled or not. If the
            Intel OpenMP Library is preloaded, this API will return True.
            Otherwise, it will return False.
    """

    return ipex._C.is_runtime_ext_enabled() == 1



import torch
try:
    import torchvision
except ImportError:
    pass  # skip if torchvision is not available

from .version import __version__
from .utils import _cpu_isa, _custom_fx_tracer
_cpu_isa.check_minimal_isa_support()

torch_version = ''
ipex_version = ''
import re
matches = re.match('(\d+\.\d+).*', torch.__version__)
if matches and len(matches.groups()) == 1:
  torch_version = matches.group(1)
matches = re.match('(\d+\.\d+).*', __version__)
if matches and len(matches.groups()) == 1:
  ipex_version = matches.group(1)
if torch_version == '' or ipex_version == '' or torch_version != ipex_version:
  print('ERROR! Intel® Extension for PyTorch* needs to work with PyTorch {0}.*, but PyTorch {1} is found. Please switch to the matching version and run again.'.format(ipex_version, torch.__version__))
  exit(127)

from . import cpu
from . import quantization
from . import nn
from . import jit
from . import profiler
from . import autocast

from .utils.verbose import verbose
from .frontend import optimize, enable_onednn_fusion, set_fp32_math_mode, get_fp32_math_mode, FP32MathMode



import torch
from torch.autograd import Function

def interaction(*args):
    r"""
    Get the interaction feature beyond different kinds of features (like gender
    or hobbies), used in DLRM model.

    For now, we only optimized "dot" interaction at `DLRM Github repo <https://github.com/facebookresearch/dlrm/blob/main/dlrm_s_pytorch.py#L475-L495>`_.
    Through this, we use the dot product to represent the interaction feature
    between two features.

    For example, if feature 1 is "Man" which is represented by [0.1, 0.2, 0.3],
    and feature 2 is "Like play football" which is represented by [-0.1, 0.3, 0.2].

    The dot interaction feature is
    ([0.1, 0.2, 0.3] * [-0.1, 0.3, 0.2]^T) =  -0.1 + 0.6 + 0.6 = 1.1

    Args:
        *args: Multiple tensors which represent different features

    Shape
        - Input: :math:`N * (B, D)`, where N is the number of different kinds of features, B is the batch size, D is feature size
        - Output: :math:`(B, D + N * ( N - 1 ) / 2)`
    """

    if torch.is_grad_enabled():
        return InteractionFunc.apply(*args)
    return torch.ops.torch_ipex.interaction_forward(args)

class InteractionFunc(Function):
    @staticmethod
    def forward(ctx, *args):
        ctx.save_for_backward(*args)
        output = torch.ops.torch_ipex.interaction_forward(args)
        return output

    @staticmethod
    def backward(ctx, grad_out):
        args = ctx.saved_tensors
        grad_in = torch.ops.torch_ipex.interaction_backward(grad_out.contiguous(), args)
        return tuple(grad_in)




from typing import List, Union

import torch
from torch import nn, Tensor
from torch.nn.modules.utils import _pair
from torch.jit.annotations import BroadcastingList2

from typing import List, Union
from .. import functional as F

class RoIAlign(nn.Module):
    """
    See :func:`roi_align`.
    """
    def __init__(
        self,
        output_size: BroadcastingList2[int],
        spatial_scale: float,
        sampling_ratio: int,
        aligned: bool = False,
    ):
        super(RoIAlign, self).__init__()
        self.output_size = output_size
        self.spatial_scale = spatial_scale
        self.sampling_ratio = sampling_ratio
        self.aligned = aligned

    def forward(self, input: Tensor, rois: Tensor) -> Tensor:
        return F._roi_align.roi_align(input, rois, self.output_size, self.spatial_scale, self.sampling_ratio, self.aligned)

    def __repr__(self) -> str:
        tmpstr = self.__class__.__name__ + '('
        tmpstr += 'output_size=' + str(self.output_size)
        tmpstr += ', spatial_scale=' + str(self.spatial_scale)
        tmpstr += ', sampling_ratio=' + str(self.sampling_ratio)
        tmpstr += ', aligned=' + str(self.aligned)
        tmpstr += ')'
        return tmpstr



import torch
import functools
import warnings
import numpy as np
import intel_extension_for_pytorch as ipex
from .cpupool import CPUPool

class Task(object):
    r"""
    An abstraction of computation based on PyTorch module and is scheduled
    asynchronously.

    Args:
        model (torch.jit.ScriptModule or torch.nn.Module): The input module.
        cpu_pool (intel_extension_for_pytorch.cpu.runtime.CPUPool): An
            intel_extension_for_pytorch.cpu.runtime.CPUPool object, contains
            all CPU cores used to run Task asynchronously.

    Returns:
        intel_extension_for_pytorch.cpu.runtime.Task: Generated
        intel_extension_for_pytorch.cpu.runtime.Task object.
    """

    def __init__(self, module, cpu_pool: CPUPool):
        self.cpu_pool = cpu_pool
        assert type(self.cpu_pool) is CPUPool
        if isinstance(module, torch.jit.ScriptModule):
            self._task = ipex._C.TaskModule(module._c, self.cpu_pool.core_ids, True)
        else:
            self._task = ipex._C.TaskModule(module, self.cpu_pool.core_ids)

    def __call__(self, *args, **kwargs):
        # async execution
        return self._task.run_async(*args, **kwargs)

    def run_sync(self, *args, **kwargs):
        # sync execution
        return self._task.run_sync(*args, **kwargs)



from .task import Task
from .cpupool import pin, CPUPool, is_runtime_ext_enabled
from .multi_stream import MultiStreamModule, get_default_num_streams, \
                        MultiStreamModuleHint, _MultiStreamBenchmarkModule
from .runtime_utils import get_core_list_of_node_id



import torch
from ._functional import _lamb_impl


class Lamb(torch.optim.Optimizer):
    r"""Implements Lamb algorithm.
    It has been proposed in `Large Batch Optimization for Deep Learning:
    Training BERT in 76 minutes`_.
    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        fused (boolean, optional): whether to use fused kernel to accelerate
            (default: False)
    .. _Large Batch Optimization for Deep Learning: Training BERT in 76 minutes:
        https://arxiv.org/abs/1904.00962
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, fused=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, fused=fused)
        super(Lamb, self).__init__(params, defaults)
        self.params_attr = {}
        self.fused = fused

    def __setstate__(self, state):
        super(Lamb, self).__setstate__(state)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.
        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avg_sqs = []
            trails = []
            state_steps = []

            for p in group['params']:
                grad = p.grad
                if grad is not None:
                    params_with_grad.append(p)
                    if grad.is_sparse:
                        raise RuntimeError('Lamb does not support sparse gradients')
                    if grad.device != torch.device('cpu'):
                        raise RuntimeError('Lamb supports only CPU device')
                    grads.append(grad)

                    state = self.state[p]
                    # Lazy state initialization
                    if len(state) == 0:
                        state['step'] = 0
                        buffer_dtype = p.dtype if p.dtype is torch.float64 else torch.float
                        state['exp_avg'] = torch.zeros(p.shape, dtype=buffer_dtype)
                        state['exp_avg_sq'] = torch.zeros(p.shape, dtype=buffer_dtype)

                    exp_avgs.append(state['exp_avg'])
                    exp_avg_sqs.append(state['exp_avg_sq'])

                    # update the steps for each param group update
                    state['step'] += 1
                    # record the step after step update
                    state_steps.append(state['step'])

            beta1, beta2 = group['betas']
            _lamb_impl(
                params_with_grad,
                grads,
                exp_avgs,
                exp_avg_sqs,
                state_steps,
                beta1,
                beta2,
                group['lr'],
                group['weight_decay'],
                group['eps'])
        return loss



import subprocess
import platform
import sys
import os

def check_minimal_isa_support():
    def get_cpu_info():
        cpu_info_path = "/proc/cpuinfo"
        if os.path.exists(cpu_info_path):
            cpu_info_command = "cat {}".format(cpu_info_path)
            all_sub_cpu_info = subprocess.getoutput(cpu_info_command).strip()
            for sub_cpu_info in all_sub_cpu_info.split("\n"):
                if sub_cpu_info.startswith("flags"):
                    cpu_flags = sub_cpu_info.replace("\t", '').upper().split(":")
                    assert len(cpu_flags) >= 2
                    all_cpu_flags = cpu_flags[1].split(" ")
                    return all_cpu_flags

            return []
        else:
            sys.exit("The extension does not support current platform - {}.".format(platform.system()))

    def check_avx2_support():
        cpu_flags = get_cpu_info()
        minimal_binary_isa = "AVX2"
        if minimal_binary_isa not in cpu_flags:
            return False

        return True

    err_msg = "ERROR! Intel® Extension for PyTorch* only works on machines with instruction sets equal or newer than AVX2, which are not detected on the current machine."
    if not check_avx2_support():
        sys.exit(err_msg)